"""
Class for finding an optimal
step alpha length in the optimization problem:
x = x + alpha * p.
It is for lcao
"""

import numpy as np
from gpaw.directmin.tools import minimum_cubic_interpol, \
    minimum_parabola_interpol


def is_descent(phi_0, phi_j, eps=1.0e-2):
    return phi_j <= phi_0 + eps * abs(phi_0)


def is_descent_and_approximate_wolfe_conditions(
        der_phi_0, phi_0, der_phi_j, phi_j, eps=1.0e-6, delta=0.1, sigma=0.9):
    """
     William W. Hager and Hongchao Zhang
     SIAM J. optim., 16(1), 170-192.
     """

    descent = is_descent(phi_0, phi_j, eps)
    awc = (2.0 * delta - 1.0) * der_phi_0 >= der_phi_j >= sigma * der_phi_0

    return descent and awc


class MaxStep(object):

    def __init__(self, evaluate_phi_and_der_phi, max_step=0.2):
        """

        :param evaluate_phi_and_der_phi:
        """

        self.evaluate_phi_and_der_phi = evaluate_phi_and_der_phi
        self.max_step = max_step
        self.name = 'max-step'

    def todict(self):
        return {'name': self.name,
                'max_step': self.max_step}

    def step_length_update(self, x, p, n_dim,
                           *args, **kwargs):

        kd = kwargs['kpdescr']
        slength = 0.0

        for k in p:
            tmp = np.linalg.norm(p[k])
            if tmp > slength:
                slength = tmp

        slength = kd.comm.max(slength)
        if slength > self.max_step:
            a_star = self.max_step / slength
        else:
            a_star = 1.0

        phi_star, der_phi_star, g_star = \
            self.evaluate_phi_and_der_phi(x, p, n_dim, a_star, *args)

        return a_star, phi_star, der_phi_star, g_star


class Parabola(MaxStep):

    """
    phi = f (x_k + a_k*p_k)
    der_phi = \\grad f(x_k + a_k p_k) \\cdot p_k
    g = \\grad f(x_k + a_k p_k)
    """

    def __init__(self, evaluate_phi_and_der_phi):
        """
        :param evaluate_phi_and_der_phi:
        """
        super(Parabola, self).__init__(evaluate_phi_and_der_phi)
        self.name = 'parabola'

    def todict(self):
        return {'name': self.name}

    def step_length_update(self, x, p, n_dim,
                           *args, **kwargs):

        phi_0 = kwargs['phi_0']
        der_phi_0 = kwargs['der_phi_0']

        phi_i, der_phi_i, g_i = \
            self.evaluate_phi_and_der_phi(x, p, n_dim, 1.0, *args)

        # if appr_wc(der_phi_0, phi_0, der_phi_i, phi_i):
        if is_descent(phi_0, phi_i, eps=1.0e-2):
            return 1.0, phi_i, der_phi_i, g_i
        else:
            a_star = minimum_parabola_interpol(
                0.0, 1.0, phi_0, phi_i, der_phi_0)
            if a_star < 0.01:
                a_star = 0.01
        phi_star, der_phi_star, g_star = \
            self.evaluate_phi_and_der_phi(x, p, n_dim,
                                          a_star, *args)

        return a_star, phi_star, der_phi_star, g_star


class StrongWolfeConditions(MaxStep):
    """
    From a book of Jorge Nocedal and Stephen J. Wright 'Numerical
    Optimization' Second Edition, 2006 (p. 56)

    This call should return a_star, phi_star, der_phi_star, g_star,
    where a_star is step length satisfied the strong Wolfe condts:

    f(x_k + a_k p_k) <= f(x_k) + c_1 a_k grad f_k cdot p_k,

    |grad f(x_k + a_k p_k) cdot p_k | <= c_2 |grad f_k cdot p_k|,

    or descent conditions and approximate Wolfe conditions from

    William W. Hager and Hongchao Zhang
    SIAM J. optim., 16(1), 170-192.

    phi = f (x_k + a_k*p_k)
    der_phi = grad f(x_k + a_k p_k) cdot p_k
    g = grad f(x_k + a_k p_k)
    """

    def __init__(self, evaluate_phi_and_der_phi,
                 c1=1.0e-4, c2=0.9,
                 searchdirtype=None, max_iter=3, eps_dx=1.0e-10,
                 eps_df=1.0e-10, use_descent_and_awc=True):
        """
        :param evaluate_phi_and_der_phi: function which calculate
        phi, der_phi and g for given A_s, P_s, n_dim and alpha
        A_s[s] is skew-hermitian matrix, P_s[s] is matrix direction
        :param searchdirtype: used only in initial guess for alpha
        :param max_iter: maximum number of iterations
        :param eps_dx: length of minimal interval where alpha can
        be found
        :param eps_df: minimal change of function
        :param c1: see above
        :param c2: see above
        :param use_descent_and_awc: check descent and approximate Wolfe
        conditions

        this class works fine, but these parameters eps_dx, eps_df
        might not be needed

        """

        super(StrongWolfeConditions, self).__init__(
            evaluate_phi_and_der_phi)

        self.max_iter = max_iter
        self.searchdirtype = searchdirtype
        self.eps_dx = eps_dx
        self.eps_df = eps_df
        self.c1 = c1
        self.c2 = c2
        self.use_descent_and_awc = use_descent_and_awc
        self.name = 'swc-awc'

    def todict(self):
        return {'name': self.name,
                'max_iter': self.max_iter,
                'searchdirtype': self.searchdirtype,
                'eps_dx': self.eps_dx,
                'eps_df': self.eps_df,
                'c1': self.c1,
                'c2': self.c2}

    def step_length_update(self, x, p, n_dim,
                           *args, **kwargs):
        c1 = self.c1
        c2 = self.c2
        phi_0 = kwargs['phi_0']
        der_phi_0 = kwargs['der_phi_0']
        phi_old = kwargs['phi_old']
        der_phi_old = kwargs['der_phi_old']
        alpha_old = kwargs['alpha_old']
        alpha_max = kwargs['alpha_max']

        alpha_1 = self.init_guess(phi_0=phi_0, der_phi_0=der_phi_0,
                                  phi_old=phi_old,
                                  der_phi_old=der_phi_old,
                                  alpha_old=alpha_old)

        i = 1
        if phi_0 is None or der_phi_0 is None:
            phi_0, der_phi_0, g_0 = \
                self.evaluate_phi_and_der_phi(x, p, n_dim, 0.0, *args)

        alpha = [0.0, alpha_1]

        phi_i_1 = phi_0
        der_phi_i_1 = der_phi_0

        max_iter = self.max_iter
        phi_max = None
        der_phi_max = None
        g_max = None

        # 'Get_step_length:'
        while True:

            if np.abs(alpha[-1] - alpha_max) < 1.0e-6 and phi_max is not None:
                phi_i, der_phi_i, g_i = phi_max, der_phi_max, g_max
                phi_max = None
                der_phi_max = None
            else:
                phi_i, der_phi_i, g_i = \
                    self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                  alpha[i],
                                                  *args)

            if self.use_descent_and_awc:
                if is_descent_and_approximate_wolfe_conditions(
                        der_phi_0, phi_0, der_phi_i, phi_i):

                    a_star = alpha[i]
                    phi_star = phi_i
                    der_phi_star = der_phi_i
                    g_star = g_i

                    break
            if phi_i > phi_0 + c1 * alpha[i] * der_phi_0 or \
                    (phi_i >= phi_i_1 and i > 1):
                a_star, phi_star, der_phi_star, g_star = \
                    self.zoom(alpha[i - 1], alpha[i],
                              phi_i_1, der_phi_i_1,
                              phi_i, der_phi_i, x, p, n_dim,
                              phi_0, der_phi_0, c1, c2, *args)
                break

            if np.fabs(der_phi_i) <= -c2 * der_phi_0:
                a_star = alpha[i]
                phi_star = phi_i
                der_phi_star = der_phi_i
                g_star = g_i
                break

            if der_phi_i >= 0.0:
                a_star, phi_star, der_phi_star, g_star = \
                    self.zoom(alpha[i], alpha[i - 1],
                              phi_i, der_phi_i,
                              phi_i_1, der_phi_i_1,
                              x, p, n_dim,
                              phi_0, der_phi_0, c1, c2, *args)
                break

            if i == max_iter:
                a_star = alpha[i]
                phi_star = phi_i
                der_phi_star = der_phi_i
                g_star = g_i

                break

            if np.abs(alpha_max - alpha[i]) < 1.0e-6:
                alpha_max = 1.5 * alpha[i]

            if phi_max is None or der_phi_max is None:
                phi_max, der_phi_max, g_max = \
                    self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                  alpha_max,
                                                  *args)

                if self.use_descent_and_awc:
                    if is_descent_and_approximate_wolfe_conditions(
                            der_phi_0, phi_0, der_phi_max, phi_max):
                        a_star = alpha_max
                        phi_star = phi_max
                        der_phi_star = der_phi_max
                        g_star = g_max
                        break

            a_new = minimum_cubic_interpol(
                alpha[i], alpha_max, phi_i, phi_max,
                der_phi_i, der_phi_max)

            alpha.append(a_new)
            phi_i_1 = phi_i
            der_phi_i_1 = der_phi_i

            if np.abs(a_new - alpha_max) < 1.0e-6:
                phi_i = phi_max
                der_phi_i = der_phi_max
                g_i = g_max

            if abs(alpha[-1] - alpha[-2]) < 1.0e-5:
                # self.log('Cannot satisfy strong Wolfe condition')
                # if i == max_iter:
                #     self.log('Too many iterations')

                a_star = alpha[i]
                phi_star = phi_i
                der_phi_star = der_phi_i
                g_star = g_i

                break

            i += 1

        return a_star, phi_star, der_phi_star, g_star

    def zoom(self, a_lo, a_hi,
             f_lo, df_lo, f_hi, df_hi, x,
             p, n_dim, phi_0, der_phi_0,
             c1, c2, *args):

        max_iter = self.max_iter
        i = 0

        while True:

            a_j = minimum_cubic_interpol(
                a_lo, a_hi, f_lo, f_hi, df_lo, df_hi)

            if a_j < 0.01:
                a_j = 0.1
                phi_j, der_phi_j, g_j = \
                    self.evaluate_phi_and_der_phi(x, p, n_dim, a_j,
                                                  *args)
                return a_j, phi_j, der_phi_j, g_j

            phi_j, der_phi_j, g_j = \
                self.evaluate_phi_and_der_phi(x, p, n_dim, a_j, *args)

            if self.use_descent_and_awc:
                if is_descent_and_approximate_wolfe_conditions(
                        der_phi_0, phi_0, der_phi_j, phi_j):

                    a_star = a_j
                    phi_star = phi_j
                    der_phi_star = der_phi_j
                    g_star = g_j

                    break

            if phi_j > phi_0 + c1 * a_j * der_phi_0 or phi_j >= f_lo:
                a_hi = a_j
                f_hi = phi_j
                df_hi = der_phi_j

            else:
                if abs(der_phi_j) <= -c2 * der_phi_0:
                    a_star = a_j
                    phi_star = phi_j
                    der_phi_star = der_phi_j
                    g_star = g_j

                    break
                if der_phi_j * (a_hi - a_lo) >= 0.0:
                    a_hi = a_lo
                    f_hi = f_lo
                    df_hi = df_lo

                a_lo = a_j
                f_lo = phi_j
                df_lo = der_phi_j

            i += 1

            if np.fabs(a_lo - a_hi) < self.eps_dx and a_lo < \
                    self.eps_dx:

                # self.log('Cannot satisfy strong Wolfe condition')
                # self.log('Interval is close to zero')
                # self.log('Lower boundary is close to zero.')

                a_star = a_lo

                phi_star, der_phi_star, g_star = \
                    self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                  a_star, *args)

                break

            elif np.fabs(a_lo - a_hi) < self.eps_dx:
                # self.log('Cannot satisfy strong Wolfe condition,')
                # self.log('Only sufficient descent')
                # self.log('Search interval is less than'
                #          '%.2e' % self.eps_dx)

                a_star = a_lo

                phi_star, der_phi_star, g_star = \
                    self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                  a_star, *args)
                break

            if i == max_iter:
                # self.log('Cannot satisfy strong Wolfe condition,')
                # self.log('Made too many iterations')
                if a_lo > self.eps_dx:
                    a_star = a_lo

                    phi_star, der_phi_star, g_star = \
                        self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                      a_star, *args)

                else:
                    a_star = a_hi

                    phi_star, der_phi_star, g_star = \
                        self.evaluate_phi_and_der_phi(x, p, n_dim,
                                                      a_star, *args)
                break

        return a_star, phi_star, der_phi_star, g_star

    def init_guess(self, phi_0, der_phi_0, phi_old, der_phi_old,
                   alpha_old=1.0):

        # chose initial alpha
        if self.searchdirtype in ['quasi-newton', 'newton']:
            alpha_1 = 1.0
        else:
            if phi_old is not None and der_phi_old is not None:
                try:
                    alpha_1 = 2.0 * (phi_0 - phi_old) / der_phi_old
                    if alpha_1 < 0.1:
                        if alpha_old < 0.1:
                            alpha_1 = 10.0
                        else:
                            alpha_1 = alpha_old
                except ZeroDivisionError:
                    alpha_1 = 1.0
            else:
                alpha_1 = 1.0

        return alpha_1



import numpy as np
import pytest
from ase.build import molecule

from gpaw import GPAW
from gpaw.lcao.projected_wannier import get_lcao_projections_HSP
from gpaw.poisson import FDPoissonSolver


@pytest.mark.later
def test_lcao_lcao_projections():
    atoms = molecule('C2H2')
    atoms.center(vacuum=3.0)
    calc = GPAW(gpts=(32, 32, 48),
                experimental={'niter_fixdensity': 2},
                poissonsolver=FDPoissonSolver(),
                eigensolver='rmm-diis')
    atoms.calc = calc
    atoms.get_potential_energy()

    V_qnM, H_qMM, S_qMM, P_aqMi = get_lcao_projections_HSP(
        calc, bfs=None, spin=0, projectionsonly=False)

    # Test H and S
    eig = sorted(np.linalg.eigvals(np.linalg.solve(S_qMM[0], H_qMM[0])).real)
    eig_ref = np.array([-17.879394403125634, -13.248793622886552,
                        -11.431247684217933, -7.125771721594344,
                        -7.125771721594252, 0.5927469818425286,
                        0.5927469818425768, 3.925124575829719,
                        7.451028687071511, 26.734351930654853])
    print(eig)
    assert np.allclose(eig, eig_ref)



import numpy as np
from scipy.linalg import eigh

from gpaw.lcao.eigensolver import DirectLCAO


class NonCollinearLDAKernel:
    name = 'LDA'
    type = 'LDA'

    def __init__(self, kernel):
        self.kernel = kernel

    def calculate(self, e_g, n_sg, v_sg):
        n_g = n_sg[0]
        m_vg = n_sg[1:4]
        m_g = (m_vg**2).sum(0)**0.5
        nnew_sg = np.empty((2,) + n_g.shape)
        nnew_sg[:] = n_g
        nnew_sg[0] += m_g
        nnew_sg[1] -= m_g
        nnew_sg *= 0.5
        vnew_sg = np.zeros_like(nnew_sg)
        self.kernel.calculate(e_g, nnew_sg, vnew_sg)
        v_sg[0] += 0.5 * vnew_sg.sum(0)
        vnew_sg /= np.where(m_g < 1e-15, 1, m_g)
        v_sg[1:4] += 0.5 * vnew_sg[0] * m_vg
        v_sg[1:4] -= 0.5 * vnew_sg[1] * m_vg


class NonCollinearLCAOEigensolver(DirectLCAO):
    def iterate(self, ham, wfs):
        wfs.timer.start('LCAO eigensolver')

        wfs.timer.start('Potential matrix')
        Vt_xdMM = [wfs.basis_functions.calculate_potential_matrices(vt_G)
                   for vt_G in ham.vt_xG]
        wfs.timer.stop('Potential matrix')

        for kpt in wfs.kpt_u:
            self.iterate_one_k_point(ham, wfs, kpt, Vt_xdMM)

        wfs.timer.stop('LCAO eigensolver')

    def iterate_one_k_point(self, ham, wfs, kpt, Vt_xdMM):
        assert wfs.gd.comm.size == 1, 'No quite sure this works!'
        if wfs.bd.comm.size > 1 and wfs.bd.strided:
            raise NotImplementedError

        H_xMM = []
        for x in range(4):
            kpt.s = x
            H_MM = self.calculate_hamiltonian_matrix(ham, wfs, kpt, Vt_xdMM[x],
                                                     root=0,
                                                     add_kinetic=(x == 0))
            H_xMM.append(H_MM)
        kpt.s = None

        S_MM = wfs.S_qMM[kpt.q]
        M = len(S_MM)
        S2_MM = np.zeros((2 * M, 2 * M), complex)
        H2_MM = np.zeros((2 * M, 2 * M), complex)

        S2_MM[:M, :M] = S_MM
        S2_MM[M:, M:] = S_MM

        # We are ignoring x and y here.
        # See gpaw.new.lcao.hamiltonian.NonCollinearHamiltonianMatrixCalculator
        # for the correct way!
        H2_MM[:M, :M] = H_xMM[0] + H_xMM[3]
        H2_MM[M:, M:] = H_xMM[0] - H_xMM[3]

        kpt.eps_n = np.empty(2 * wfs.bd.mynbands)

        diagonalization_string = repr(self.diagonalizer)
        wfs.timer.start(diagonalization_string)
        kpt.eps_n, V_MM = eigh(H2_MM, S2_MM)
        kpt.C_nM = V_MM.T.conj()
        wfs.timer.stop(diagonalization_string)
        kpt.C_nM.shape = (wfs.bd.mynbands * 4, M)



from ase.units import Bohr
from gpaw.new.ase_interface import GPAW
from gpaw.utilities.dipole import dipole_matrix_elements_from_calc


def test_dipole_me(gpw_files):
    """Check dipole matrix-elements for H2 molecule."""
    calc = GPAW(gpw_files['h2_pw_wfs'])

    # Method 1: evaluate all-electron wave functions on fine grid:
    psi0, psi1 = (
        calc.calculation.state.ibzwfs.get_all_electron_wave_function(n)
        for n in [0, 1])
    if psi0 is not None:
        d1_v = (psi0 * psi1).moment() * Bohr

    # Method 2: use pseudo wave function + PAW corrections:
    d2_nnv = dipole_matrix_elements_from_calc(calc, n1=0, n2=2)[0]

    assert abs(d2_nnv[0, 0] - calc.atoms.cell.sum(0) / 2).max() < 0.04
    assert abs(d2_nnv[1, 1] - calc.atoms.cell.sum(0) / 2).max() < 0.04
    if psi0 is not None:
        assert abs(d2_nnv[0, 1] - d1_v).max() < 1e-3

    # Method 3: same as above but with translated molecule:
    calc = GPAW(gpw_files['h2_pw_0_wfs'])
    d3_nnv = dipole_matrix_elements_from_calc(calc, n1=0, n2=2,
                                              center=[0, 0, 0])[0]

    assert abs(d3_nnv[0, 0]).max() < 0.04
    assert abs(d3_nnv[1, 1]).max() < 0.04
    assert abs(abs(d3_nnv[0, 1]) - abs(d2_nnv[0, 1])).max() < 1e-6



from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.mixer import BroydenMixer
from gpaw.test import equal


def test_broydenmixer(in_tmp_dir):
    a = 2.7
    bulk = Atoms([Atom('Li')], pbc=True, cell=(a, a, a))
    k = 2
    g = 16
    calc = GPAW(gpts=(g, g, g), kpts=(k, k, k), nbands=2,
                mixer=BroydenMixer())
    bulk.calc = calc
    e = bulk.get_potential_energy()
    calc.write('Li.gpw')
    GPAW('Li.gpw')

    energy_tolerance = 0.0001
    equal(e, -1.20258, energy_tolerance)



"""Test all electron density for right interpretation of coreholes"""
import pytest
from ase.build import molecule
from ase.units import Bohr
from gpaw import GPAW, PoissonSolver
from gpaw.mixer import Mixer
from gpaw.test import gen


@pytest.mark.later
def test_aed_with_corehole_li():
    """Compare number of electrons for different channels with corehole"""
    li_setup = gen('Li', name='fch1s', corehole=(1, 0, 1), xcname='PBE')
    grf = 1
    atoms = molecule('Li2')
    atoms.center(vacuum=2.5)

    calc = GPAW(xc='PBE',
                mixer=Mixer(),
                setups={0: li_setup},
                charge=-1,
                poissonsolver=PoissonSolver('fd'))
    atoms.calc = calc
    atoms.get_potential_energy()

    n_sg = calc.get_all_electron_density(gridrefinement=grf)

    ne_sz = calc.density.gd.integrate(
        n_sg, global_integral=False) * (Bohr / grf)**3
    assert ne_sz == pytest.approx(6.0, abs=1e-5)

    atoms.set_initial_magnetic_moments([0.66, .34])
    calc = calc.new(spinpol=True)
    atoms.calc = calc
    atoms.get_potential_energy()

    for sz in range(2):
        n_sg = calc.get_all_electron_density(spin=sz, gridrefinement=grf)
        ne_sz = calc.density.gd.integrate(
            n_sg, global_integral=False) * (Bohr / grf)**3
        assert ne_sz == pytest.approx(3.0, abs=1e-5)



from gpaw.fd_operators import Gradient, FDOperator
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.mpi import world


def test_fd_ops_gradient():
    if world.size > 4:
        # Grid is so small that domain decomposition cannot exceed 4 domains
        assert world.size % 4 == 0
        group, other = divmod(world.rank, 4)
        ranks = np.arange(4 * group, 4 * (group + 1))
        domain_comm = world.new_communicator(ranks)
    else:
        domain_comm = world

    gd = GridDescriptor((8, 1, 1), (8.0, 1.0, 1.0), comm=domain_comm)
    a = gd.zeros()
    dadx = gd.zeros()
    a[:, 0, 0] = np.arange(gd.beg_c[0], gd.end_c[0])
    gradx = Gradient(gd, v=0)
    print(a.itemsize, a.dtype, a.shape)
    print(dadx.itemsize, dadx.dtype, dadx.shape)
    gradx.apply(a, dadx)

    #   a = [ 0.  1.  2.  3.  4.  5.  6.  7.]

    dAdx = gd.collect(dadx, broadcast=True)
    assert not (dAdx[:, 0, 0] - [-3, 1, 1, 1, 1, 1, 1, -3]).any()

    # Backwards FD operator:
    gradx2 = FDOperator([-1, 1], [[-1, 0, 0], [0, 0, 0]], gd)
    gradx2.apply(a, dadx)
    dAdx = gd.collect(dadx, broadcast=True)
    assert not (dAdx[:, 0, 0] - [-7, 1, 1, 1, 1, 1, 1, 1]).any()

    gd = GridDescriptor((1, 8, 1), (1.0, 8.0, 1.0),
                        (1, 0, 1), comm=domain_comm)
    dady = gd.zeros()
    a = gd.zeros()
    grady = Gradient(gd, v=1)
    a[0, :, 0] = np.arange(gd.beg_c[1], gd.end_c[1]) - 1
    grady.apply(a, dady)

    #   da
    #   -- = [0.5  1.   1.   1.   1.   1.  -2.5]
    #   dy

    dady = gd.collect(dady, broadcast=True)
    assert dady[0, 0, 0] == 0.5 and np.sum(dady[0, :, 0]) == 3.0

    # a GUC cell
    gd = GridDescriptor((1, 7, 1),
                        ((1.0, 0.0, 0.0),
                         (5.0, 5.0, 0.0),
                         (0.0, 0.0, 0.7)), comm=domain_comm)
    dady = gd.zeros()
    grady = Gradient(gd, v=1)
    a = gd.zeros()
    a[0, :, 0] = np.arange(gd.beg_c[1], gd.end_c[1]) - 1
    grady.apply(a, dady)

    #   da
    #   -- = [-3.5  1.4   1.4   1.4   1.4   1.4  -3.5]
    #   dy

    dady = gd.collect(dady, broadcast=True)
    assert abs(dady[0, 0, 0] - -
               3.5) < 1e-12 and abs(np.sum(dady[0, :, 0])) < 1e-12

    # Check continuity of weights:
    weights = []
    for x in np.linspace(-0.6, 0.6, 130, True):
        gd = GridDescriptor((3, 3, 3),
                            ((3.0, 0.0, 0.0),
                             (3 * x, 1.5 * 3**0.5, 0.0),
                             (0.0, 0.0, 3)),
                            comm=domain_comm)
        g = Gradient(gd, v=0)
        for c, o in zip(g.coef_p, g.offset_pc):
            if (o == (-1, 1, 0)).all():
                break
        else:
            c = 0.0
        weights.append(c)

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(weights)
        plt.show()



from math import sqrt
from ase import Atoms
from gpaw import GPAW, FermiDirac


def test_symmetry_usesymm2():
    a = 2.4437
    # Set up graphene structure
    vacuum = 3
    atoms = Atoms(symbols='C2',
                  positions=[(0.5 * a, -sqrt(3) / 6 * a, vacuum),
                             (0.5 * a, +sqrt(3) / 6 * a, vacuum)],
                  cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                        (0.5 * a, +0.5 * 3**0.5 * a, 0),
                        (0.0, 0.0, 2 * vacuum)])
    atoms.pbc = True

    # Gamma-point calculation of graphene
    calc = GPAW(h=0.2,
                occupations=FermiDirac(width=0.15),
                kpts=(1, 1, 1),
                xc='LDA')
    atoms.calc = calc
    atoms.get_potential_energy()

    kpts = [(1 / 2.0, 1 / 3.0, 0)]

    # Calculate one K-point with symmetry:
    calc = calc.fixed_density(kpts=kpts)
    eigs_True = calc.get_eigenvalues(kpt=0)

    # Redo with the same K-point without symmetry:
    calc = calc.fixed_density(kpts=kpts,
                              symmetry='off')
    eigs_False = calc.get_eigenvalues(kpt=0)

    assert abs(eigs_True[0] - eigs_False[0]) < 1e-4



import numpy as np
from .radialgd import RadialGridDescriptor


def shape_functions(rgd: RadialGridDescriptor,
                    type: str,
                    rc: float,
                    lmax: int) -> np.ndarray:
    """Shape functions for compensation charges."""
    g_lg = rgd.zeros(lmax + 1)
    r_g = rgd.r_g

    if type == 'gauss':
        g_lg[0] = 4 / rc**3 / np.sqrt(np.pi) * np.exp(-(r_g / rc)**2)
        for l in range(1, lmax + 1):
            g_lg[l] = 2.0 / (2 * l + 1) / rc**2 * r_g * g_lg[l - 1]
    elif type == 'sinc':
        g_lg[0] = np.sinc(r_g / rc)**2
        g_lg[0, rgd.ceil(rc):] = 0.0
        for l in range(1, lmax + 1):
            g_lg[l] = r_g * g_lg[l - 1]
    elif type == 'bessel':
        from scipy.special import spherical_jn as jn
        roots = [[3.141592653589793, 6.283185307179586],
                 [4.493409457909095, 7.7252518369375],
                 [5.76345919689455, 9.095011330476355]]
        for l in range(lmax + 1):
            q1, q2 = (x0 / rc for x0 in roots[l])
            alpha = -q1 / q2 * jn(l, q1 * rc, True) / jn(l, q2 * rc, True)
            g_lg[l] = jn(l, q1 * r_g) + alpha * jn(l, q2 * r_g)
        g_lg[:, rgd.ceil(rc):] = 0.0
    else:
        1 / 0

    for l in range(lmax + 1):
        g_lg[l] /= rgd.integrate(g_lg[l], l) / (4 * np.pi)

    return g_lg


if __name__ == '__main__':
    from .radialgd import EquidistantRadialGridDescriptor as RGD
    from scipy.special import spherical_jn as jn
    from scipy.optimize import root
    import matplotlib.pyplot as plt

    r = np.linspace(0, 1.2, 200)

    if 0:
        # Find roots of spherical Bessel functions:
        for l in range(3):
            for i, x0 in enumerate([3, 6]):
                result = root(lambda x: jn(l, x), x0 + 1.5 * l)
                x0 = result['x']
                print(l, i, x0.item())
                plt.plot(r, jn(l, r * x0), label=str(x0))
        plt.legend()
        plt.show()

    rgd = RGD(0.01, 120)
    for l in range(3):
        rc = 0.3
        for type in ['gauss', 'sinc', 'bessel']:
            g_lg = shape_functions(rgd, type, rc, l)
            plt.plot(rgd.r_g, g_lg[l], label=type)  # type: ignore
            rc = 1.0

        plt.legend()
        plt.show()



"""
==  ==========
R
r
G
g
h
x   r or h
==  ==========

"""

from __future__ import annotations

from collections import defaultdict
from typing import DefaultDict

import numpy as np
from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.new import zip
from gpaw.new.potential import Potential
from gpaw.new.xc import XCFunctional
from gpaw.setup import Setup
from gpaw.spinorbit import soc as soc_terms
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import pack, pack2, unpack
from gpaw.yml import indent
from gpaw.mpi import serial_comm


class PotentialCalculator:
    def __init__(self,
                 xc: XCFunctional,
                 poisson_solver,
                 setups: list[Setup],
                 nct_R: UniformGridFunctions,
                 fracpos_ac: Array2D,
                 soc: bool = False):
        self.poisson_solver = poisson_solver
        self.xc = xc
        self.setups = setups
        self.nct_R = nct_R
        self.fracpos_ac = fracpos_ac
        self.soc = soc

    def __str__(self):
        return (f'{self.poisson_solver}\n'
                f'xc functional:\n{indent(self.xc)}\n')

    def calculate_pseudo_potential(self,
                                   density,
                                   vHt_x: DistributedArrays | None
                                   ) -> tuple[dict[str, float],
                                              UniformGridFunctions,
                                              DistributedArrays]:
        raise NotImplementedError

    def calculate_charges(self, vHt_x):
        raise NotImplementedError

    def calculate(self,
                  density,
                  vHt_x: DistributedArrays | None = None,
                  kpt_comm=serial_comm
                  ) -> tuple[Potential, DistributedArrays, AtomArrays]:
        energies, vt_sR, vHt_x = self.calculate_pseudo_potential(
            density, vHt_x)

        Q_aL = self.calculate_charges(vHt_x)
        dH_asii, corrections = calculate_non_local_potential(
            self.setups, density, self.xc, Q_aL, self.soc, kpt_comm)

        for key, e in corrections.items():
            # print(f'{key:10} {energies[key]:15.9f} {e:15.9f}')
            energies[key] += e

        return Potential(vt_sR, dH_asii, energies), vHt_x, Q_aL

    def move(self, fracpos_ac, atomdist, ndensities) -> UniformGridFunctions:
        """Move things and return change in pseudo core density."""
        delta_nct_R = self.nct_R.new()
        delta_nct_R.data[:] = self.nct_R.data
        delta_nct_R.data *= -1
        self._move(fracpos_ac, atomdist, ndensities)
        delta_nct_R.data += self.nct_R.data
        return delta_nct_R

    def _move(self, fracpos_ac, atomdist, ndensities) -> None:
        raise NotImplementedError


def calculate_non_local_potential(setups,
                                  density,
                                  xc,
                                  Q_aL,
                                  soc: bool,
                                  comm) -> tuple[AtomArrays,
                                                 dict[str, float]]:
    dtype = float if density.ncomponents < 4 else complex
    D_asii = density.D_asii.to_xp(np)
    dH_asii = D_asii.layout.new(dtype=dtype).empty(density.ncomponents)
    Q_aL = Q_aL.to_xp(np)
    energy_corrections: DefaultDict[str, float] = defaultdict(float)
    for a, D_sii in D_asii.items():
        if a % comm.size != comm.rank:
            dH_asii[a][:] = 0.0
            continue
        Q_L = Q_aL[a]
        setup = setups[a]
        dH_sii, corrections = calculate_non_local_potential1(
            setup, xc, D_sii, Q_L, soc)
        dH_asii[a][:] = dH_sii
        for key, e in corrections.items():
            energy_corrections[key] += e

    comm.sum(dH_asii.data)
    # Sum over domain:
    names = ['kinetic', 'coulomb', 'zero', 'xc', 'external']
    energies = np.array([energy_corrections[name] for name in names])
    density.D_asii.layout.atomdist.comm.sum(energies)
    comm.sum(energies)

    return (dH_asii.to_xp(density.D_asii.layout.xp),
            {name: e for name, e in zip(names, energies)})


def calculate_non_local_potential1(setup: Setup,
                                   xc: XCFunctional,
                                   D_sii: Array3D,
                                   Q_L: Array1D,
                                   soc: bool) -> tuple[Array3D,
                                                       dict[str, float]]:
    ncomponents = len(D_sii)
    ndensities = 2 if ncomponents == 2 else 1
    D_sp = np.array([pack(D_ii) for D_ii in D_sii])

    D_p = D_sp[:ndensities].sum(0)

    dH_p = (setup.K_p + setup.M_p +
            setup.MB_p + 2.0 * setup.M_pp @ D_p +
            setup.Delta_pL @ Q_L)
    e_kinetic = setup.K_p @ D_p + setup.Kc
    e_zero = setup.MB + setup.MB_p @ D_p
    e_coulomb = setup.M + D_p @ (setup.M_p + setup.M_pp @ D_p)

    dH_sp = np.zeros_like(D_sp, dtype=float if ncomponents < 4 else complex)
    if soc:
        dH_sp[1:4] = pack2(soc_terms(setup, xc.xc, D_sp))
    dH_sp[:ndensities] = dH_p
    e_xc = xc.calculate_paw_correction(setup, D_sp, dH_sp)
    e_kinetic -= (D_sp * dH_sp).sum().real
    e_external = 0.0

    dH_sii = unpack(dH_sp)

    return dH_sii, {'kinetic': e_kinetic,
                    'coulomb': e_coulomb,
                    'zero': e_zero,
                    'xc': e_xc,
                    'external': e_external}



# for n > 4 (?)
# when run with gpaw-python built with MKL matrices b and c have ~zero elements
# running with python (uses _gpaw.so) gives right (non-zero) results!

import numpy as np
from gpaw.utilities.blas import mmm


def test_linalg_gemm_complex():
    n = 5  # works for n = 4
    a1 = np.eye(n) + 1.j
    a2 = a1 + 1.j

    b = np.zeros((n, n), dtype=complex)

    c = np.dot(a2, a1)

    mmm(1.0, a2, 'N', a1, 'N', 0.0, b)

    thresh = 1.0e-7
    ref_max_value = -9.0

    # print b
    # print c
    numpy_dot = np.max(b).real
    gpaw_gemm = np.max(c).real
    # print gpaw_gemm
    assert abs(gpaw_gemm - numpy_dot) < thresh, (gpaw_gemm, numpy_dot, thresh)
    assert abs(gpaw_gemm - ref_max_value) < thresh, (gpaw_gemm,
                                                     ref_max_value, thresh)



import time

import pytest

import gpaw.fftw as fftw

N = 1
# n1, n2, n3 = 32, 28, 128
n1, n2, n3 = 8, 6, 12


def check(Plan, flags, input, output, sign):
    t0 = time.time()
    plan = Plan(input, output, sign, flags)
    t1 = time.time()
    t = 0.0
    for i in range(N):
        input[:] = 1.3
        t2 = time.time()
        plan.execute()
        t3 = time.time()
        t += t3 - t2
    return t1 - t0, t / N


@pytest.mark.skipif(not fftw.have_fftw(), reason='No FFTW')
def test_fft():
    a1 = fftw.empty((n1, n2, n3), complex)
    a2 = fftw.empty((n1, n2, n3), complex)
    b = fftw.empty((n1, n2, n3 // 2 + 1), complex)
    c1 = b.view(dtype=float)[:, :, :n3]
    c2 = fftw.empty((n1, n2, n3 // 2), complex).view(dtype=float)
    for input, output, sign in [
        (a1, a1, -1),
        (a1, a2, -1),
        (b, c1, 1),
        (b, c2, 1),
        (c1, b, -1),
        (c2, b, -1)]:
        for Plan, flags in [(fftw.NumpyFFTPlan, 117),
                            (fftw.FFTWPlan, fftw.ESTIMATE),
                            (fftw.FFTWPlan, fftw.MEASURE),
                            (fftw.FFTWPlan, fftw.PATIENT),
                            (fftw.FFTWPlan, fftw.EXHAUSTIVE)]:
            tplan, tfft = check(Plan, flags, input, output, sign)
            print(('%-12s %3d %10.6f %10.6f' %
                  (Plan.__name__, flags, tplan, tfft)))



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Grid-descriptors

This module contains a classes defining uniform 3D grids.
For radial grid descriptors, look atom/radialgd.py.

"""

import numbers
from math import pi
from typing import Sequence

import numpy as np
from scipy.ndimage import map_coordinates

import _gpaw
import gpaw.mpi as mpi
from gpaw.domain import Domain
from gpaw.new import prod
from gpaw.typing import Array1D, Array3D, Vector
from gpaw.utilities.blas import mmm, r2k, rk

NONBLOCKING = False


class GridBoundsError(ValueError):
    pass


class BadGridError(ValueError):
    pass


class GridDescriptor(Domain):
    r"""Descriptor-class for uniform 3D grid

    A ``GridDescriptor`` object holds information on how functions, such
    as wave functions and electron densities, are discreticed in a
    certain domain in space.  The main information here is how many
    grid points are used in each direction of the unit cell.

    There are methods for tasks such as allocating arrays, performing
    symmetry operations and integrating functions over space.  All
    methods work correctly also when the domain is parallelized via
    domain decomposition.

    This is how a 2x2x2 3D array is laid out in memory::

        3-----7
        |\    |\
        | \   | \
        |  1-----5      z
        2--|--6  |   y  |
         \ |   \ |    \ |
          \|    \|     \|
           0-----4      +-----x

    Example:

     >>> a = np.zeros((2, 2, 2))
     >>> a.ravel()[:] = range(8)
     >>> a
     array([[[0., 1.],
             [2., 3.]],
     <BLANKLINE>
            [[4., 5.],
             [6., 7.]]])
     """

    ndim = 3  # dimension of ndarrays

    def __init__(self, N_c, cell_cv=[1, 1, 1], pbc_c=True,
                 comm=None, parsize_c=None, allow_empty_domains=False):
        """Construct grid-descriptor object.

        parameters:

        N_c: 3 ints
            Number of grid points along axes.
        cell_cv: 3 float's or 3x3 floats
            Unit cell.
        pbc_c: one or three bools
            Periodic boundary conditions flag(s).
        comm: MPI-communicator
            Communicator for domain-decomposition.
        parsize_c: tuple of 3 ints, a single int or None
            Number of domains.
        allow_empty_domains: bool
            Allow parallelization that would generate empty domains.

        Note that if pbc_c[c] is False, then the actual number of gridpoints
        along axis c is one less than N_c[c].

        Attributes:

        ==========  ========================================================
        ``dv``      Volume per grid point.
        ``h_cv``    Array of the grid spacing along the three axes.
        ``N_c``     Array of the number of grid points along the three axes.
        ``n_c``     Number of grid points on this CPU.
        ``beg_c``   Beginning of grid-point indices (inclusive).
        ``end_c``   End of grid-point indices (exclusive).
        ``comm``    MPI-communicator for domain decomposition.
        ==========  ========================================================

        The length unit is Bohr.
        """

        if isinstance(pbc_c, int):
            pbc_c = (pbc_c,) * 3
        if comm is None:
            comm = mpi.world

        self.N_c = np.array(N_c, int)
        if (self.N_c != N_c).any():
            raise ValueError('Non-int number of grid points %s' % N_c)

        Domain.__init__(self, cell_cv, pbc_c, comm, parsize_c, self.N_c)
        self.rank = self.comm.rank

        self.beg_c = np.empty(3, int)
        self.end_c = np.empty(3, int)

        self.n_cp = []
        for c in range(3):
            n_p = (np.arange(self.parsize_c[c] + 1) * float(self.N_c[c]) /
                   self.parsize_c[c])
            n_p = np.around(n_p + 0.4999).astype(int)

            if not self.pbc_c[c]:
                n_p[0] = 1

            if np.any(n_p[1:] == n_p[:-1]):
                if allow_empty_domains:
                    # If there are empty domains, sort them to the end
                    n_p[:] = (np.arange(self.parsize_c[c] + 1) +
                              1 - self.pbc_c[c]).clip(0, self.N_c[c])
                else:
                    msg = ('Grid {0} too small for {1} cores!'
                           .format('x'.join(str(n) for n in self.N_c),
                                   'x'.join(str(n) for n in self.parsize_c)))
                    raise BadGridError(msg)

            self.beg_c[c] = n_p[self.parpos_c[c]]
            self.end_c[c] = n_p[self.parpos_c[c] + 1]
            self.n_cp.append(n_p)

        self.n_c = self.end_c - self.beg_c

        self.h_cv = self.cell_cv / self.N_c[:, np.newaxis]
        self.volume = abs(np.linalg.det(self.cell_cv))
        self.dv = self.volume / self.N_c.prod()

        self.orthogonal = not (self.cell_cv -
                               np.diag(self.cell_cv.diagonal())).any()

    def __repr__(self):
        if self.orthogonal:
            cellstring = np.diag(self.cell_cv).tolist()
        else:
            cellstring = self.cell_cv.tolist()

        pcoords = tuple(self.get_processor_position_from_rank())
        return ('GridDescriptor(%s, cell_cv=%s, pbc_c=%s, comm=[%d/%d, '
                'domain=%s], parsize=%s)'
                % (self.N_c.tolist(), cellstring,
                   np.array(self.pbc_c).astype(int).tolist(), self.comm.rank,
                   self.comm.size, pcoords, self.parsize_c.tolist()))

    def new_descriptor(self, N_c=None, cell_cv=None, pbc_c=None,
                       comm=None, parsize_c=None, allow_empty_domains=False):
        """Create new descriptor based on this one.

        The new descriptor will use the same class (possibly a subclass)
        and all arguments will be equal to those of this descriptor
        unless new arguments are provided."""
        if N_c is None:
            N_c = self.N_c
        if cell_cv is None:
            cell_cv = self.cell_cv
        if pbc_c is None:
            pbc_c = self.pbc_c
        if comm is None:
            comm = self.comm
        if parsize_c is None and comm.size == self.comm.size:
            parsize_c = self.parsize_c
        return self.__class__(N_c, cell_cv, pbc_c, comm, parsize_c,
                              allow_empty_domains)

    def coords(self, c, pad=True):
        """Return coordinates along one of the three axes.

        Useful for plotting::

            import matplotlib.pyplot as plt
            plt.plot(gd.coords(0), data[:, 0, 0])
            plt.show()

        """
        L = np.linalg.norm(self.cell_cv[c])
        N = self.N_c[c]
        h = L / N
        p = self.pbc_c[c] or pad
        return np.linspace((1 - p) * h, L, N - 1 + p, False)

    def get_grid_spacings(self):
        L_c = (np.linalg.inv(self.cell_cv)**2).sum(0)**-0.5
        return L_c / self.N_c

    def get_size_of_global_array(self, pad=False):
        if pad:
            return self.N_c
        else:
            return self.N_c - 1 + self.pbc_c

    def flat_index(self, G_c):
        g1, g2, g3 = G_c - self.beg_c
        return g3 + self.n_c[2] * (g2 + g1 * self.n_c[1])

    def get_slice(self):
        return [slice(b - 1 + p, e - 1 + p) for b, e, p in
                zip(self.beg_c, self.end_c, self.pbc_c)]

    def zeros(self, n=(), dtype=float, global_array=False, pad=False):
        """Return new zeroed 3D array for this domain.

        The type can be set with the ``dtype`` keyword (default:
        ``float``).  Extra dimensions can be added with ``n=dim``.  A
        global array spanning all domains can be allocated with
        ``global_array=True``."""

        return self._new_array(n, dtype, True, global_array, pad)

    def empty(self, n=(), dtype=float, global_array=False, pad=False):
        """Return new uninitialized 3D array for this domain.

        The type can be set with the ``dtype`` keyword (default:
        ``float``).  Extra dimensions can be added with ``n=dim``.  A
        global array spanning all domains can be allocated with
        ``global_array=True``."""

        return self._new_array(n, dtype, False, global_array, pad)

    def _new_array(self, n=(), dtype=float, zero=True,
                   global_array=False, pad=False):
        if global_array:
            shape = self.get_size_of_global_array(pad)
        else:
            shape = self.n_c

        if isinstance(n, numbers.Integral):
            n = (n,)

        shape = n + tuple(shape)

        if zero:
            return np.zeros(shape, dtype)
        else:
            return np.empty(shape, dtype)

    def get_axial_communicator(self, axis):
        peer_ranks = []
        pos_c = self.parpos_c.copy()
        for i in range(self.parsize_c[axis]):
            pos_c[axis] = i
            peer_ranks.append(self.get_rank_from_processor_position(pos_c))
        peer_comm = self.comm.new_communicator(peer_ranks)
        return peer_comm

    def integrate(self, a_xg, b_yg=None,
                  global_integral=True, hermitian=False):
        """Integrate function(s) over domain.

        a_xg: ndarray
            Function(s) to be integrated.
        b_yg: ndarray
            If present, integrate a_xg.conj() * b_yg.
        global_integral: bool
            If the array(s) are distributed over several domains, then the
            total sum will be returned.  To get the local contribution
            only, use global_integral=False.
        hermitian: bool
            Result is hermitian.
        """

        xshape = a_xg.shape[:-3]

        if b_yg is None:
            # Only one array:
            result = a_xg.reshape(xshape + (-1,)).sum(axis=-1) * self.dv
            if global_integral:
                if result.ndim == 0:
                    result = self.comm.sum(result)
                else:
                    self.comm.sum(result)
            return result

        gsize = prod(a_xg.shape[-3:])
        A_xg = np.ascontiguousarray(a_xg.reshape((-1, gsize)))
        B_yg = np.ascontiguousarray(b_yg.reshape((-1, gsize)))

        result_yx = np.zeros((len(B_yg), len(A_xg)), A_xg.dtype)

        if a_xg is b_yg:
            rk(self.dv, A_xg, 0.0, result_yx)
        elif hermitian:
            r2k(0.5 * self.dv, A_xg, B_yg, 0.0, result_yx)
        else:
            # gemm(self.dv, A_xg, B_yg, 0.0, result_yx, 'c')
            mmm(self.dv, B_yg, 'N', A_xg, 'C', 0.0, result_yx)

        if global_integral:
            self.comm.sum(result_yx)

        yshape = b_yg.shape[:-3]
        result = result_yx.T.reshape(xshape + yshape)

        if result.ndim == 0:
            return result.item()
        else:
            return result

    def coarsen(self):
        """Return coarsened `GridDescriptor` object.

        Reurned descriptor has 2x2x2 fewer grid points."""

        if np.sometrue(self.N_c % 2):
            raise ValueError('Grid %s not divisible by 2!' % self.N_c)

        return self.new_descriptor(self.N_c // 2)

    def refine(self):
        """Return refined `GridDescriptor` object.

        Returned descriptor has 2x2x2 more grid points."""
        return self.new_descriptor(self.N_c * 2)

    def get_boxes(self, spos_c, rcut, cut=True):
        """Find boxes enclosing sphere."""
        N_c = self.N_c
        ncut = rcut * (self.icell_cv**2).sum(axis=1)**0.5 * self.N_c
        npos_c = spos_c * N_c
        beg_c = np.ceil(npos_c - ncut).astype(int)
        end_c = np.ceil(npos_c + ncut).astype(int)

        if cut:
            for c in range(3):
                if not self.pbc_c[c]:
                    if beg_c[c] < 0:
                        beg_c[c] = 0
                    if end_c[c] > N_c[c]:
                        end_c[c] = N_c[c]
        else:
            for c in range(3):
                if (not self.pbc_c[c] and
                    (beg_c[c] < 0 or end_c[c] > N_c[c])):
                    msg = ('Box at %.3f %.3f %.3f crosses boundary.  '
                           'Beg. of box %s, end of box %s, max box size %s' %
                           (tuple(spos_c) + (beg_c, end_c, self.N_c)))
                    raise GridBoundsError(msg)

        range_c = ([], [], [])

        for c in range(3):
            b = beg_c[c]
            e = b

            while e < end_c[c]:
                b0 = b % N_c[c]

                e = min(end_c[c], b + N_c[c] - b0)

                if b0 < self.beg_c[c]:
                    b1 = b + self.beg_c[c] - b0
                else:
                    b1 = b

                e0 = b0 - b + e

                if e0 > self.end_c[c]:
                    e1 = e - (e0 - self.end_c[c])
                else:
                    e1 = e
                if e1 > b1:
                    range_c[c].append((b1, e1))
                b = e

        boxes = []

        for b0, e0 in range_c[0]:
            for b1, e1 in range_c[1]:
                for b2, e2 in range_c[2]:
                    b = np.array((b0, b1, b2))
                    e = np.array((e0, e1, e2))
                    beg_c = np.array((b0 % N_c[0], b1 % N_c[1], b2 % N_c[2]))
                    end_c = beg_c + e - b
                    disp = (b - beg_c) / N_c
                    beg_c = np.maximum(beg_c, self.beg_c)
                    end_c = np.minimum(end_c, self.end_c)
                    if (beg_c[0] < end_c[0] and
                        beg_c[1] < end_c[1] and
                        beg_c[2] < end_c[2]):
                        boxes.append((beg_c, end_c, disp))

        return boxes

    def get_nearest_grid_point(self, spos_c, force_to_this_domain=False):
        """Return index of nearest grid point.

        The nearest grid point can be on a different CPU than the one the
        nucleus belongs to (i.e. return can be negative, or larger than
        gd.end_c), in which case something clever should be done.
        The point can be forced to the grid descriptors domain to be
        consistent with self.get_rank_from_position(spos_c).
        """
        g_c = np.around(self.N_c * spos_c).astype(int)
        if force_to_this_domain:
            for c in range(3):
                g_c[c] = max(g_c[c], self.beg_c[c])
                g_c[c] = min(g_c[c], self.end_c[c] - 1)
        return g_c - self.beg_c

    def plane_wave(self, k_c):
        """Evaluate plane wave on grid.

        Returns::

               _ _
              ik.r
             e    ,

        where the wave vector is given by k_c (in units of reciprocal
        lattice vectors)."""

        index_Gc = np.indices(self.n_c).T + self.beg_c
        return np.exp(2j * pi * np.dot(index_Gc, k_c / self.N_c).T)

    def symmetrize(self, a_g, op_scc, ft_sc=None):
        # ft_sc: fractional translations
        # XXXX documentation missing.  This is some kind of array then?
        if len(op_scc) == 1:
            return

        if ft_sc is not None and not ft_sc.any():
            ft_sc = None

        A_g = self.collect(a_g)
        if self.comm.rank == 0:
            B_g = np.zeros_like(A_g)
            for s, op_cc in enumerate(op_scc):
                if ft_sc is None:
                    _gpaw.symmetrize(A_g, B_g, op_cc, 1 - self.pbc_c)
                else:
                    t_c = (ft_sc[s] * self.N_c).round().astype(int)
                    _gpaw.symmetrize_ft(A_g, B_g, op_cc, t_c,
                                        1 - self.pbc_c)
        else:
            B_g = None
        self.distribute(B_g, a_g)
        a_g /= len(op_scc)

    def collect(self, a_xg, out=None, broadcast=False):
        """Collect distributed array to master-CPU or all CPU's."""
        if self.comm.size == 1:
            if out is None:
                return a_xg
            out[:] = a_xg
            return out

        xshape = a_xg.shape[:-3]

        # Collect all arrays on the master:
        if self.rank != 0:
            # There can be several sends before the corresponding receives
            # are posted, so use syncronous send here
            self.comm.ssend(a_xg, 0, 301)
            if broadcast:
                A_xg = self.empty(xshape, a_xg.dtype, global_array=True)
                self.comm.broadcast(A_xg, 0)
                return A_xg
            else:
                return np.nan

        # Put the subdomains from the slaves into the big array
        # for the whole domain:
        if out is None:
            A_xg = self.empty(xshape, a_xg.dtype, global_array=True)
        else:
            A_xg = out
        parsize_c = self.parsize_c
        r = 0
        for n0 in range(parsize_c[0]):
            b0, e0 = self.n_cp[0][n0:n0 + 2] - self.beg_c[0]
            for n1 in range(parsize_c[1]):
                b1, e1 = self.n_cp[1][n1:n1 + 2] - self.beg_c[1]
                for n2 in range(parsize_c[2]):
                    b2, e2 = self.n_cp[2][n2:n2 + 2] - self.beg_c[2]
                    if r != 0:
                        a_xg = np.empty(xshape +
                                        ((e0 - b0), (e1 - b1), (e2 - b2)),
                                        a_xg.dtype.char)
                        self.comm.receive(a_xg, r, 301)
                    A_xg[..., b0:e0, b1:e1, b2:e2] = a_xg
                    r += 1
        if broadcast:
            self.comm.broadcast(A_xg, 0)
        return A_xg

    def distribute(self, B_xg, out=None):
        """Distribute full array B_xg to subdomains, result in b_xg.

        B_xg is not used by the slaves (i.e. it should be None on all slaves)
        b_xg must be allocated on all nodes and will be overwritten.
        """

        if self.comm.size == 1:
            if out is None:
                return B_xg
            out[:] = B_xg
            return out

        if out is None:
            out = self.empty(B_xg.shape[:-3], dtype=B_xg.dtype)

        if self.rank != 0:
            self.comm.receive(out, 0, 42)
        else:
            parsize_c = self.parsize_c
            requests = []
            r = 0
            for n0 in range(parsize_c[0]):
                b0, e0 = self.n_cp[0][n0:n0 + 2] - self.beg_c[0]
                for n1 in range(parsize_c[1]):
                    b1, e1 = self.n_cp[1][n1:n1 + 2] - self.beg_c[1]
                    for n2 in range(parsize_c[2]):
                        b2, e2 = self.n_cp[2][n2:n2 + 2] - self.beg_c[2]
                        if r != 0:
                            a_xg = B_xg[..., b0:e0, b1:e1, b2:e2].copy()
                            request = self.comm.send(a_xg, r, 42, NONBLOCKING)
                            # Remember to store a reference to the
                            # send buffer (a_xg) so that is isn't
                            # deallocated:
                            requests.append((request, a_xg))
                        else:
                            out[:] = B_xg[..., b0:e0, b1:e1, b2:e2]
                        r += 1

            for request, a_xg in requests:
                self.comm.wait(request)

        return out

    def zero_pad(self, a_xg, global_array=True):
        """Pad array with zeros as first element along non-periodic directions.

        Array may either be local or in standard decomposition.
        """

        # We could infer what global_array should be from a_xg.shape.
        # But as it is now, there is a bit of redundancy to avoid
        # confusing errors

        gshape = a_xg.shape[-3:]
        padding_c = 1 - self.pbc_c
        if global_array:
            assert (gshape == self.N_c - padding_c).all(), gshape
            bshape = tuple(self.N_c)
        else:
            assert (gshape == self.n_c).all()
            parpos_c = self.get_processor_position_from_rank()
            # Only pad where domain is on edge:
            padding_c *= (parpos_c == 0)
            bshape = tuple(self.n_c + padding_c)

        if self.pbc_c.all():
            return a_xg

        npbx, npby, npbz = padding_c
        b_xg = np.zeros(a_xg.shape[:-3] + tuple(bshape), dtype=a_xg.dtype)
        b_xg[..., npbx:, npby:, npbz:] = a_xg
        return b_xg

    def dipole_moment(self,
                      rho_R: Array3D,
                      center_v: Vector = None) -> Array1D:
        """Calculate dipole moment of density.

        Integration region will be centered on center_v.  Default center
        is center of unit cell.
        """
        index_cr = [np.arange(self.beg_c[c], self.end_c[c], dtype=float)
                    for c in range(3)]

        if center_v is not None:
            corner_c = (np.linalg.solve(self.h_cv.T,
                                        center_v) % self.N_c) - self.N_c / 2
            for corner, index_r, N in zip(corner_c, index_cr, self.N_c):
                index_r -= corner
                index_r %= N
                index_r += corner

        rho_ijk = rho_R
        rho_ij = rho_ijk.sum(axis=2)
        rho_ik = rho_ijk.sum(axis=1)
        rho_cr = [rho_ij.sum(axis=1), rho_ij.sum(axis=0), rho_ik.sum(axis=0)]

        d_c = [np.dot(index_cr[c], rho_cr[c]) for c in range(3)]
        d_v = -np.dot(d_c, self.h_cv) * self.dv
        self.comm.sum(d_v)
        return d_v

    def calculate_dipole_moment(self, rho_g, center=False, origin_c=None):
        """Calculate dipole moment of density."""
        r_cz = [np.arange(self.beg_c[c], self.end_c[c]) for c in range(3)]
        if center:
            assert origin_c is None
            r_cz = [r_cz[c] - 0.5 * self.N_c[c] for c in range(3)]
        elif origin_c is not None:
            r_cz = [r_cz[c] - origin_c[c] for c in range(3)]

        rho_01 = rho_g.sum(axis=2)
        rho_02 = rho_g.sum(axis=1)
        rho_cz = [rho_01.sum(axis=1), rho_01.sum(axis=0), rho_02.sum(axis=0)]
        rhog_c = [np.dot(r_cz[c], rho_cz[c]) for c in range(3)]
        d_c = -np.dot(rhog_c, self.h_cv) * self.dv
        self.comm.sum(d_c)
        return d_c

    def wannier_matrix(self, psit_nG, psit_nG1, G_c, nbands=None):
        """Wannier localization integrals

        The soft part of Z is given by (Eq. 27 ref1)::

            ~       ~     -i G.r   ~
            Z   = <psi | e      |psi >
             nm       n             m

        psit_nG and psit_nG1 are the set of wave functions for the two
        different spin/kpoints in question.

        ref1: Thygesen et al, Phys. Rev. B 72, 125119 (2005)
        """

        if nbands is None:
            nbands = len(psit_nG)

        if nbands == 0:
            return np.zeros((0, 0), complex)

        e_G = np.exp(-2j * pi * np.dot(np.indices(self.n_c).T +
                                       self.beg_c, G_c / self.N_c).T)
        a_nG = (e_G * psit_nG[:nbands].conj()).reshape((nbands, -1))
        return np.inner(a_nG,
                        psit_nG1[:nbands].reshape((nbands, -1))) * self.dv

    def find_center(self, a_R):
        """Calculate center of positive function."""
        assert self.orthogonal
        r_vR = self.get_grid_point_coordinates()
        a_R = a_R.astype(complex)
        center = []
        for L, r_R in zip(self.cell_cv.diagonal(), r_vR):
            z = self.integrate(a_R, np.exp(2j * pi / L * r_R))
            center.append(np.angle(z) / (2 * pi) * L % L)
        return np.array(center)

    def bytecount(self, dtype=float):
        """Get the number of bytes used by a grid of specified dtype."""
        return int(np.prod(self.n_c)) * np.array(1, dtype).itemsize

    def get_grid_point_coordinates(self, dtype=float, global_array=False):
        """Construct cartesian coordinates of grid points in the domain."""
        r_vG = np.dot(np.indices(self.n_c, dtype).T + self.beg_c,
                      self.h_cv).T.copy()
        if global_array:
            return self.collect(r_vG, broadcast=True)  # XXX waste!
        else:
            return r_vG

    def get_grid_point_distance_vectors(self, r_v, mic=True, dtype=float):
        """Return distances to a given vector in the domain.

        mic: if true adopts the mininimum image convention
        procedure by W. Smith in 'The Minimum image convention in
        Non-Cubic MD cells' March 29, 1989
        """
        s_Gc = (np.indices(self.n_c, dtype).T + self.beg_c) / self.N_c
        r_c = np.linalg.solve(self.cell_cv.T, r_v)
        # do the correction twice works better because of rounding errors
        # e.g.: -1.56250000e-25 % 1.0 = 1.0,
        #      but (-1.56250000e-25 % 1.0) % 1.0 = 0.0
        r_c = np.where(self.pbc_c, r_c % 1.0, r_c)
        s_Gc -= np.where(self.pbc_c, r_c % 1.0, r_c)

        if mic:
            s_Gc -= self.pbc_c * (2 * s_Gc).astype(int)
            # sanity check
            assert (s_Gc * self.pbc_c >= -0.5).all()
            assert (s_Gc * self.pbc_c <= 0.5).all()

        return np.dot(s_Gc, self.cell_cv).T.copy()

    def interpolate_grid_points(self, spos_nc, vt_g):
        """Return interpolated values.

        Calculate interpolated values from array vt_g based on the
        scaled coordinates on spos_c.

        This doesn't work in parallel, since it would require
        communication between neighbouring grids."""

        assert self.comm.size == 1

        vt_g = self.zero_pad(vt_g)
        return map_coordinates(vt_g,
                               (spos_nc * self.N_c).T,
                               order=3,
                               mode='wrap')

    def is_my_grid_point(self, R_c: Sequence[int]) -> bool:
        """Check if grid point belongs to this domain."""
        return ((self.beg_c <= R_c) & (R_c < self.end_c)).all()



def frozen(cls):
    """Make sure attributes do not appear out of nowhere.

    From:

        https://stackoverflow.com/questions/3603502/
        prevent-creating-new-attributes-outside-init
    """
    def frozensetattr(self, key, value):
        import inspect
        if not hasattr(self, key) and inspect.stack()[1][3] != '__init__':
            raise AttributeError('Class {} is frozen. Cannot set {}'
                                 .format(cls.__name__, key))
        else:
            self.__dict__[key] = value

    cls.__setattr__ = frozensetattr
    return cls



import pytest
from ase.build import molecule
from gpaw import GPAW, Davidson, Mixer, PoissonSolver
from gpaw.utilities.tools import split_formula
from gpaw.test import equal


@pytest.mark.mgga
def test_tpss(in_tmp_dir):
    cell = [6., 6., 7.]

    # Reference from J. Chem. Phys. Vol 120 No. 15, 15 April 2004, page 6898
    tpss_de = {
        'Li2': 22.5}
    tpss_old = {
        'Li2': 20.9}

    exp_bonds_dE = {
        'Li2': (2.673, 24.4)}

    systems = ['Li2']

    # Add atoms
    for formula in systems:
        temp = split_formula(formula)
        for atom in temp:
            if atom not in systems:
                systems.append(atom)
    energies = {}

    # Calculate energies
    for formula in systems:
        loa = molecule(formula)
        loa.set_cell(cell)
        loa.center()
        calc = GPAW(h=0.3,
                    eigensolver=Davidson(8),
                    parallel=dict(kpt=1),
                    mixer=Mixer(0.5, 5),
                    nbands=-2,
                    poissonsolver=PoissonSolver('fd', relax='GS'),
                    xc='oldPBE',
                    txt=formula + '.txt')
        if len(loa) == 1:
            calc = calc.new(hund=True)
        else:
            pos = loa.get_positions()
            pos[1, :] = pos[0, :] + [0.0, 0.0, exp_bonds_dE[formula][0]]
            loa.set_positions(pos)
            loa.center()
        loa.calc = calc
        energy = loa.get_potential_energy()
        diff = calc.get_xc_difference('TPSS')
        energies[formula] = (energy, energy + diff)
        print(formula, energy, energy + diff)

    # calculate atomization energies
    print('formula\tGPAW\tRef\tGPAW-Ref\tGPAW-exp')
    mae_ref, mae_exp, mae_pbe, count = 0.0, 0.0, 0.0, 0
    for formula in tpss_de.keys():
        atoms_formula = split_formula(formula)
        de_tpss = -1.0 * energies[formula][1]
        de_pbe = -1.0 * energies[formula][0]
        for atom_formula in atoms_formula:
            de_tpss += energies[atom_formula][1]
            de_pbe += energies[atom_formula][0]

        de_tpss *= 627.5 / 27.211
        de_pbe *= 627.5 / 27.211
        mae_ref += abs(de_tpss - tpss_de[formula])
        mae_exp += abs(de_tpss - exp_bonds_dE[formula][1])
        mae_pbe += abs(de_pbe - exp_bonds_dE[formula][1])
        count += 1
        out = ("%s\t%.1f\t%.1f\t%.1f\t%.1f kcal/mol" %
               (formula, de_tpss, tpss_de[formula],
                de_tpss - tpss_de[formula],
                de_tpss - exp_bonds_dE[formula][1]))
        print(out)

        # comparison to gpaw revision 5450 version value
        # in kcal/mol (note the grid:0.3 Ang)
        equal(de_tpss, tpss_old[formula], 0.15)



# Copyright (C) 2008  CSC Scientific Computing Ltd.
# Please see the accompanying LICENSE file for further information.

"""This module defines an Overlap operator.

The module defines an overlap operator and implements overlap-related
functions.

"""
import numpy as np


class OverlapCorrections:
    def __init__(self, setups):
        self.setups = setups

    def apply(self, P, out=None):
        if out is None:
            out = P.new()
        for a, I1, I2 in P.indices:
            dS_ii = self.setups[a].dO_ii
            out.array[..., I1:I2] = np.dot(P.array[..., I1:I2], dS_ii)
        return out


class Overlap:
    """Overlap operator S

    This class contains information required to apply the
    overlap operator to a set of wavefunctions.
    """

    def __init__(self, timer=None):
        """Create the Overlap operator."""
        self.timer = timer

    def apply(self, a_xG, b_xG, wfs, kpt, calculate_P_ani=True):
        """Apply the overlap operator to a set of vectors.

        Parameters
        ==========
        a_nG: ndarray
            Set of vectors to which the overlap operator is applied.
        b_nG: ndarray, output
            Resulting S times a_nG vectors.
        kpt: KPoint object
            k-point object defined in kpoint.py.
        calculate_P_ani: bool
            When True, the integrals of projector times vectors
            P_ni = <p_i | a_nG> are calculated.
            When False, existing P_ani are used

        """
        b_xG[:] = a_xG
        shape = a_xG.shape[:-3]
        P_axi = wfs.pt.dict(shape)

        if calculate_P_ani:
            wfs.pt.integrate(a_xG, P_axi, kpt.q)
        else:
            for a, P_ni in kpt.P_ani.items():
                P_axi[a][:] = P_ni

        for a, P_xi in P_axi.items():
            P_axi[a] = np.dot(P_xi, wfs.setups[a].dO_ii)
            # gemm(1.0, wfs.setups[a].dO_ii, P_xi, 0.0, P_xi, 'n')
        wfs.pt.add(b_xG, P_axi, kpt.q)  # b_xG += sum_ai pt^a_i P_axi

    def apply_inverse(self, a_xG, b_xG, wfs, kpt, calculate_P_ani=True):
        """Apply approximative inverse overlap operator to wave functions."""

        b_xG[:] = a_xG
        shape = a_xG.shape[:-3]
        P_axi = wfs.pt.dict(shape)

        if calculate_P_ani:
            wfs.pt.integrate(a_xG, P_axi, kpt.q)
        else:
            for a, P_ni in kpt.P_ani.items():
                P_axi[a][:] = P_ni

        for a, P_xi in P_axi.items():
            P_axi[a] = np.dot(P_xi, wfs.setups[a].dC_ii)
        wfs.pt.add(b_xG, P_axi, kpt.q)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

# FIXME: This code used to be used in dfpt/phononcalculator.py in replacement
#        for the original FFTPoissonSolver. This code should be dead.

from math import pi

from numpy.fft import fftn, ifftn

from gpaw.utilities.tools import construct_reciprocal
from gpaw import poisson


class PoissonSolver(poisson.FDPoissonSolver):
    """Copy of GPAW Poissonsolver"""


class FFTPoissonSolver(poisson.FFTPoissonSolver):

    def __init__(self, eps=2e-10, dtype=float):
        """Set the ``dtype`` of the source term array."""

        self.dtype = dtype
        self.eps = eps

    def set_q(self, q_c):
        """Set q-vector in case of Bloch-type charge distribution.

        Parameters
        ----------
        q_c: ndarray
            q-vector in scaled coordinates of the reciprocal lattice vectors.

        """

        if self.gd.comm.rank == 0:
            if self.gd.comm.size > 1:
                raise RuntimeError(
                    'This solver is obsolete:' +
                    'domain decomposition support has been removed')
            self.k2_Q, self.N3 = construct_reciprocal(self.gd, q_c=q_c)

    def solve_neutral(self, phi_g, rho_g, eps=None):
        """Solve Poissons equation for a neutral and periodic charge density.

        Parameters
        ----------
        phi_g: ndarray
            Potential (output array).
        rho_g: ndarray
            Charge distribution (in units of -e).

        """

        assert phi_g.dtype == self.dtype
        assert rho_g.dtype == self.dtype

        if self.gd.comm.size == 1:
            # Note, implicit downcast from complex to float when the dtype of
            # phi_g is float
            phi_g[:] = ifftn(fftn(rho_g) * 4.0 * pi / self.k2_Q)
        else:
            rho_g = self.gd.collect(rho_g)
            if self.gd.comm.rank == 0:
                globalphi_g = ifftn(fftn(rho_g) * 4.0 * pi / self.k2_Q)
            else:
                globalphi_g = None
            # What happens here if globalphi is complex and phi is real ??????
            self.gd.distribute(globalphi_g, phi_g)

        return 1



import pytest
from ase import Atoms
from gpaw import GPAW, PW


@pytest.mark.ci
def test_pw_smallanglecell(in_tmp_dir):
    a = 3.0
    ec = 200
    h2 = Atoms('H2', [[0, 0, 0], [0, 0, 0.8]],
               cell=[a, a, a], pbc=1)
    h2.calc = GPAW(mode=PW(ec), txt='sc.txt')
    e0 = h2.get_potential_energy()
    h2.cell[1, 0] = a
    with pytest.warns(UserWarning):
        e = h2.get_potential_energy()
    assert abs(e - e0) < 0.001, e - e0



"""
Search directions in space of skew-hermitian matrices

LSR1 algorithm and application to excited states:
arXiv:2006.15922 [physics.chem-ph]
J. Chem. Theory Comput. 16, 6968 (2020).
https://doi.org/10.1021/acs.jctc.0c00597
"""

import numpy as np
import copy
from gpaw.directmin.tools import array_to_dict, dict_to_array


class SearchDirectionBase(object):
    """
    Base class for search direction algorithms
    """

    def __init__(self):
        self.iters = 0
        self.kp = None
        self.p = None
        self.k = None
        super(SearchDirectionBase, self).__init__()

    def __str__(self):
        raise NotImplementedError('Search direction class needs string '
                                  'representation.')

    def todict(self):
        raise NotImplementedError('Search direction class needs \'todict\' '
                                  'method.')

    def update_data(self, wfs, x_k1, g_k1, precond=None):
        raise NotImplementedError('Search direction class needs '
                                  '\'update_data\' method.')

    def reset(self):
        self.iters = 0
        self.kp = {}
        self.p = 0
        self.k = 0


class ModeFollowingBase(object):
    """
    Base gradient partitioning and negation implementation for minimum mode
    following
    """

    def __init__(self, partial_diagonalizer, convex_step_length=0.1,
                 reset_on_convex=False):
        self.eigv = None
        self.eigvec = None
        self.eigvec_old = None
        self.partial_diagonalizer = partial_diagonalizer
        self.fixed_sp_order = None
        self.was_convex = False
        self.convex_step_length = convex_step_length
        self.reset_on_convex = reset_on_convex

    def update_eigenpairs(self, g_k1, wfs, ham, dens):
        """
        Performs a partial Hessian diagonalization to obtain the eigenvectors
        with negative eigenvalues.

        :param g_k1: Gradient.
        :param wfs:
        :param ham:
        :param dens:
        :return:
        """

        self.partial_diagonalizer.grad = g_k1
        use_prev = False if self.eigv is None or (
            self.was_convex and self.reset_on_convex) else True
        self.partial_diagonalizer.run(wfs, ham, dens, use_prev)
        self.eigv = copy.deepcopy(self.partial_diagonalizer.lambda_w)
        self.eigvec_old = copy.deepcopy(self.eigvec)
        self.eigvec = copy.deepcopy(self.partial_diagonalizer.x_w.T)
        if wfs.dtype == complex:
            dimtot = int(len(self.eigvec[0]) / 2)
            eigvec = np.zeros(shape=(len(self.eigvec), dimtot),
                              dtype=complex)
            for i in range(len(self.eigvec)):
                eigvec[i] += self.eigvec[i][: dimtot] \
                    + 1.0j * self.eigvec[i][dimtot:]
            self.eigvec = eigvec
        self.fixed_sp_order = self.partial_diagonalizer.sp_order

    def invert_parallel_grad(self, g_k1):
        """
        Uses the stored eigenpairs and inverts the projections of the gradient
        parallel to the eigenvectors with negative eigenvalues.

        :param g_k1: Gradient.
        :return: Modified gradient.
        """

        grad, dim, dimtot = dict_to_array(g_k1)
        get_dots = 0
        if self.fixed_sp_order is None:
            for i in range(len(self.eigv)):
                if self.eigv[i] <= -1e-4:
                    get_dots += 1
                else:
                    break
        else:
            neg_temp = 0
            for i in range(len(self.eigv)):
                if self.eigv[i] <= -1e-4:
                    neg_temp += 1
                else:
                    break
            get_dots = self.fixed_sp_order
        grad_par = np.zeros_like(grad)
        if self.fixed_sp_order is not None:
            for i in range(get_dots):
                grad_par += self.eigvec[i] * np.dot(self.eigvec[i].conj(),
                                                    grad.T).real
            grad_mod = grad - 2.0 * grad_par
        else:
            for i in range(get_dots):
                grad_par += self.eigvec[i] \
                    * np.dot(self.eigvec[i].conj(), grad.T).real
            if get_dots == 0:
                grad_mod = -self.convex_step_length * grad_par \
                    / np.linalg.norm(grad_par)
                self.partial_diagonalizer.etdm.searchdir_algo.reset()
                self.was_convex = True
            else:
                grad_mod = grad - 2.0 * grad_par
                if self.was_convex:
                    self.partial_diagonalizer.etdm.searchdir_algo.reset()
                    self.was_convex = False
        return array_to_dict(grad_mod, dim)


class ModeFollowing(ModeFollowingBase, SearchDirectionBase):
    """
    Minimum mode following class handling the GMF tag of the search direction
    class for ETDM and negation of the gradient projection.
    """

    def __init__(self, partial_diagonalizer, search_direction,
                 convex_step_length=0.1):
        self.sd = search_direction
        self.name = self.sd.name + '_gmf'
        self.type = self.sd.type + '_gmf'
        super(ModeFollowing, self).__init__(partial_diagonalizer,
                                            convex_step_length)

    @property
    def beta_0(self):
        return self.sd.beta_0

    def __str__(self):
        return self.sd.__str__() + ' with minimum mode following'

    def todict(self):
        res = self.sd.todict()
        res['name'] += '_gmf'                    # tag will be removed in etdm
        return res

    def update_data(self, wfs, x_k1, g_k1, precond=None):
        g_k1 = self.invert_parallel_grad(g_k1)
        return self.sd.update_data(wfs, x_k1, g_k1, precond=precond)


class SteepestDescent(SearchDirectionBase):
    """
    Steepest descent algorithm
    """

    def __init__(self):
        super(SteepestDescent, self).__init__()

        self.name = 'sd'
        self.type = 'steepest-descent'

    def __str__(self):
        return 'Steepest Descent algorithm'

    def todict(self):
        return {'name': self.name}

    def update_data(self, wfs, x_k1, g_k1, precond=None):

        if precond is None:
            p_k = minus(g_k1)
        else:
            p_k = apply_prec(precond, g_k1, -1.0)
        self.iters += 1
        return p_k


class FRcg(SteepestDescent):
    """
    The Fletcher-Reeves conj. grad. method
    See Jorge Nocedal and Stephen J. Wright 'Numerical
    Optimization' Second Edition, 2006 (p. 121)
    """

    def __init__(self):
        super(FRcg, self).__init__()
        self.name = 'fr-cg'
        self.type = 'conjugate-gradients'

    def __str__(self):
        return 'Fletcher-Reeves conjugate gradient method'

    def todict(self):
        return {'name': self.name}

    def update_data(self, wfs, x_k1, g_k1, precond=None):

        if precond is not None:
            g_k1 = apply_prec(precond, g_k1, 1.0)

        if self.iters == 0:
            self.p_k = minus(g_k1)
        else:
            dot_g_k1_g_k1 = dot_all_k_and_b(g_k1, g_k1, wfs)
            dot_g_g = dot_all_k_and_b(self.g_k, self.g_k, wfs)
            beta_k = dot_g_k1_g_k1 / dot_g_g
            self.p_k = calc_diff(self.p_k, g_k1, beta_k)

        # save this step
        self.g_k = copy.deepcopy(g_k1)

        self.iters += 1
        return self.p_k


class QuickMin(SearchDirectionBase):

    def __init__(self, dt=0.01, mass=0.01):
        super(QuickMin, self).__init__()
        self.dt = dt
        self.mass = mass
        self.name = 'quick-min'
        self.type = 'equation-of-motion'

    def __str__(self):

        return 'QuickMin'

    def todict(self):
        return {'name': self.name,
                'dt': self.dt,
                'mass': self.mass}

    def update_data(self, wfs, x_k1, g_k1, precond=None):

        if precond is not None:
            g_k1 = apply_prec(precond, g_k1, 1.0)

        dt = self.dt
        m = self.mass

        if self.iters == 0:
            self.v = multiply(g_k1, -dt / m)
            p = multiply(self.v, dt)
        else:
            dot_gv = dot_all_k_and_b(g_k1, self.v, wfs)
            dot_gg = dot_all_k_and_b(g_k1, g_k1, wfs)
            if dot_gv > 0.0:
                dot_gv = 0.0
            gamma = (dt / m - dot_gv / dot_gg)
            self.v = multiply(g_k1, -gamma)
            p = multiply(self.v, dt)

        self.iters += 1
        return p


class LBFGS(SearchDirectionBase):

    def __init__(self, memory=3):
        """
        :param memory: amount of previous steps to use
        """
        super(LBFGS, self).__init__()

        self.s_k = {i: None for i in range(memory)}
        self.y_k = {i: None for i in range(memory)}

        self.rho_k = np.zeros(shape=memory)

        self.kp = {}
        self.p = 0
        self.k = 0

        self.memory = memory
        self.stable = True
        self.name = 'l-bfgs'
        self.type = 'quasi-newton'

    def __str__(self):

        return 'L-BFGS'

    def todict(self):
        return {'name': self.name,
                'memory': self.memory}

    def update_data(self, wfs, x_k1, g_k1, precond=None):

        self.iters += 1

        if precond is not None:
            g_k1 = apply_prec(precond, g_k1, 1.0)

        if self.k == 0:

            self.kp[self.k] = self.p
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = g_k1
            self.s_k[self.kp[self.k]] = zeros(g_k1)
            self.y_k[self.kp[self.k]] = zeros(g_k1)
            self.k += 1
            self.p += 1
            self.kp[self.k] = self.p

            return minus(g_k1)

        else:

            if self.p == self.memory:
                self.p = 0
                self.kp[self.k] = self.p

            s_k = self.s_k
            x_k = self.x_k
            y_k = self.y_k
            g_k = self.g_k

            x_k1 = copy.deepcopy(x_k1)

            rho_k = self.rho_k

            kp = self.kp
            k = self.k
            m = self.memory

            s_k[kp[k]] = calc_diff(x_k1, x_k)
            y_k[kp[k]] = calc_diff(g_k1, g_k)
            dot_ys = dot_all_k_and_b(y_k[kp[k]], s_k[kp[k]], wfs)

            if abs(dot_ys) > 1.0e-15:
                rho_k[kp[k]] = 1.0 / dot_ys
            else:
                rho_k[kp[k]] = 1.0e15

            if dot_ys < 0.0:
                self.stable = False

            q = copy.deepcopy(g_k1)

            alpha = np.zeros(np.minimum(k + 1, m))
            j = np.maximum(-1, k - m)

            for i in range(k, j, -1):
                dot_sq = dot_all_k_and_b(s_k[kp[i]], q, wfs)

                alpha[kp[i]] = rho_k[kp[i]] * dot_sq

                q = calc_diff(q, y_k[kp[i]], const=alpha[kp[i]])

            t = k
            dot_yy = dot_all_k_and_b(y_k[kp[t]], y_k[kp[t]], wfs)

            if abs(dot_yy) > 1.0e-15:
                r = multiply(q, 1.0 / (rho_k[kp[t]] * dot_yy))
            else:
                r = multiply(q, 1.0e15)

            for i in range(np.maximum(0, k - m + 1), k + 1):
                dot_yr = dot_all_k_and_b(y_k[kp[i]], r, wfs)

                beta = rho_k[kp[i]] * dot_yr

                r = calc_diff(r, s_k[kp[i]], const=(beta - alpha[kp[i]]))

            # save this step:
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = copy.deepcopy(g_k1)
            self.k += 1
            self.p += 1
            self.kp[self.k] = self.p

            return multiply(r, -1.0)


class LBFGS_P(SearchDirectionBase):

    def __init__(self, memory=3, beta_0=1.0):
        """
        :param memory: amount of previous steps to use
        """
        super(LBFGS_P, self).__init__()
        self.s_k = {i: None for i in range(memory)}
        self.y_k = {i: None for i in range(memory)}
        self.rho_k = np.zeros(shape=memory)
        self.kp = {}
        self.p = 0
        self.k = 0
        self.memory = memory
        self.stable = True
        self.beta_0 = beta_0
        self.name = 'l-bfgs-p'
        self.type = 'quasi-newton'

    def __str__(self):

        return 'L-BFGS-P'

    def todict(self):

        return {'name': self.name,
                'memory': self.memory,
                'beta_0': self.beta_0}

    def update_data(self, wfs, x_k1, g_k1, precond=None):
        # For L-BFGS-P, the preconditioner passed here has to be differentiated
        # from the preconditioner passed in ETDM. To keep the UI of this member
        # function consistent, the term precond is still used in the signature
        hess_1 = precond
        self.iters += 1
        if self.k == 0:
            self.kp[self.k] = self.p
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = copy.deepcopy(g_k1)
            self.s_k[self.kp[self.k]] = zeros(g_k1)
            self.y_k[self.kp[self.k]] = zeros(g_k1)
            self.k += 1
            self.p += 1
            self.kp[self.k] = self.p
            if hess_1 is None:
                p = minus(g_k1)
            else:
                p = apply_prec(hess_1, g_k1, -1.0)
            self.beta_0 = 1.0
            return p

        else:
            if self.p == self.memory:
                self.p = 0
                self.kp[self.k] = self.p

            s_k = self.s_k
            x_k = self.x_k
            y_k = self.y_k
            g_k = self.g_k
            rho_k = self.rho_k
            kp = self.kp
            k = self.k
            m = self.memory

            s_k[kp[k]] = calc_diff(x_k1, x_k)
            y_k[kp[k]] = calc_diff(g_k1, g_k)

            dot_ys = dot_all_k_and_b(y_k[kp[k]], s_k[kp[k]], wfs)

            if abs(dot_ys) > 1.0e-20:
                rho_k[kp[k]] = 1.0 / dot_ys
            else:
                rho_k[kp[k]] = 1.0e20

            if rho_k[kp[k]] < 0.0:
                self.stable = False
                self.__init__(memory=self.memory)
                return self.update_data(wfs, x_k1, g_k1, hess_1)

            q = copy.deepcopy(g_k1)

            alpha = np.zeros(np.minimum(k + 1, m))
            j = np.maximum(-1, k - m)

            for i in range(k, j, -1):
                dot_sq = dot_all_k_and_b(s_k[kp[i]], q, wfs)
                alpha[kp[i]] = rho_k[kp[i]] * dot_sq
                q = calc_diff(q, y_k[kp[i]], const=alpha[kp[i]])

            t = k
            dot_yy = dot_all_k_and_b(y_k[kp[t]], y_k[kp[t]], wfs)
            rhoyy = rho_k[kp[t]] * dot_yy
            if abs(rhoyy) > 1.0e-20:
                self.beta_0 = 1.0 / rhoyy
            else:
                self.beta_0 = 1.0e20

            if hess_1 is not None:
                r = apply_prec(hess_1, q)
            else:
                r = multiply(q, self.beta_0)

            for i in range(np.maximum(0, k - m + 1), k + 1):
                dot_yr = dot_all_k_and_b(y_k[kp[i]], r, wfs)
                beta = rho_k[kp[i]] * dot_yr
                r = calc_diff(r, s_k[kp[i]], const=(beta - alpha[kp[i]]))

            # save this step:
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = copy.deepcopy(g_k1)

            self.k += 1
            self.p += 1

            self.kp[self.k] = self.p

            return multiply(r, -1.0)


class LSR1P(SearchDirectionBase):
    """
    This class describes limited memory versions of
    SR-1, Powell and their combintaions (such as Bofill).
    """

    def __init__(self, memory=20, method='LSR1', phi=None):
        """
        :param memory: amount of previous steps to use
        """
        super(LSR1P, self).__init__()

        self.u_k = {i: None for i in range(memory)}
        self.j_k = {i: None for i in range(memory)}
        self.yj_k = np.zeros(shape=memory)
        self.method = method
        self.phi = phi

        self.phi_k = np.zeros(shape=memory)
        if self.phi is None:
            assert self.method in ['LSR1', 'LP', 'LBofill',
                                   'Linverse_Bofill'], 'Value Error'
            if self.method == 'LP':
                self.phi_k.fill(1.0)
        else:
            self.phi_k.fill(self.phi)

        self.kp = {}
        self.p = 0
        self.k = 0

        self.memory = memory
        self.name = 'l-sr1p'
        self.type = 'quasi-newton'

    def __str__(self):

        return 'LSR1P'

    def todict(self):

        return {'name': self.name,
                'memory': self.memory,
                'method': self.method}

    def update_data(self, wfs, x_k1, g_k1, precond=None):

        if precond is not None:
            bg_k1 = apply_prec(precond, g_k1, 1.0)
        else:
            bg_k1 = g_k1.copy()

        if self.k == 0:
            self.kp[self.k] = self.p
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = copy.deepcopy(g_k1)
            self.u_k[self.kp[self.k]] = zeros(g_k1)
            self.j_k[self.kp[self.k]] = zeros(g_k1)
            self.k += 1
            self.p += 1
            self.kp[self.k] = self.p
        else:
            if self.p == self.memory:
                self.p = 0
                self.kp[self.k] = self.p

            x_k = self.x_k
            g_k = self.g_k
            u_k = self.u_k
            j_k = self.j_k
            yj_k = self.yj_k
            phi_k = self.phi_k

            x_k1 = copy.deepcopy(x_k1)

            kp = self.kp
            k = self.k
            m = self.memory

            s_k = calc_diff(x_k1, x_k)
            y_k = calc_diff(g_k1, g_k)
            if precond is not None:
                by_k = apply_prec(precond, y_k, 1.0)
            else:
                by_k = y_k.copy()

            by_k = self.update_bv(wfs, by_k, y_k, u_k, j_k, yj_k, phi_k,
                                  np.maximum(1, k - m), k)

            j_k[kp[k]] = calc_diff(s_k, by_k)
            yj_k[kp[k]] = dot_all_k_and_b(y_k, j_k[kp[k]], wfs)

            if self.method == 'LSR1':
                if abs(yj_k[kp[k]]) < 1e-12:
                    yj_k[kp[k]] = 1e-12

            dot_yy = dot_all_k_and_b(y_k, y_k, wfs)
            if abs(dot_yy) > 1.0e-15:
                u_k[kp[k]] = multiply(y_k, 1.0 / dot_yy)
            else:
                u_k[kp[k]] = multiply(y_k, 1.0e15)

            if self.method == 'LBofill' and self.phi is None:
                jj_k = dot_all_k_and_b(j_k[kp[k]], j_k[kp[k]], wfs)
                phi_k[kp[k]] = 1 - yj_k[kp[k]]**2 / (dot_yy * jj_k)
            elif self.method == 'Linverse_Bofill' and self.phi is None:
                jj_k = dot_all_k_and_b(j_k[kp[k]], j_k[kp[k]], wfs)
                phi_k[kp[k]] = yj_k[kp[k]] ** 2 / (dot_yy * jj_k)

            bg_k1 = self.update_bv(wfs, bg_k1, g_k1, u_k, j_k, yj_k, phi_k,
                                   np.maximum(1, k - m + 1), k + 1)

            # save this step:
            self.x_k = copy.deepcopy(x_k1)
            self.g_k = copy.deepcopy(g_k1)
            self.k += 1
            self.p += 1
            self.kp[self.k] = self.p

        self.iters += 1
        return multiply(bg_k1, -1.0)

    def update_bv(self, wfs, bv, v, u_k, j_k, yj_k, phi_k, i_0, i_m):
        kp = self.kp

        for i in range(i_0, i_m):
            dot_uv = dot_all_k_and_b(u_k[kp[i]], v, wfs)
            dot_jv = dot_all_k_and_b(j_k[kp[i]], v, wfs)

            alpha = dot_jv - yj_k[kp[i]] * dot_uv
            beta_p = calc_diff(j_k[kp[i]], u_k[kp[i]], dot_uv, -alpha)

            beta_ms = multiply(j_k[kp[i]], dot_jv / yj_k[kp[i]])

            beta = calc_diff(beta_ms, beta_p, 1 - phi_k[kp[i]], -phi_k[kp[i]])

            bv = calc_diff(bv, beta, const=-1.0)

        return bv


def multiply(x, const=1.0):
    """
    it must not change x!
    :param x:
    :param const:
    :return: new dictionary y = cons*x
    """
    y = {}
    for k in x:
        y[k] = const * x[k]
    return y


def zeros(x):
    y = {}
    for k in x.keys():
        y[k] = np.zeros_like(x[k])
    return y


def minus(x):
    return multiply(x, -1.0)


def calc_diff(x1, x2, const_0=1.0, const=1.0):
    y_k = {}
    for k in x1.keys():
        y_k[k] = const_0 * x1[k] - const * x2[k]
    return y_k


def dot_all_k_and_b(x1, x2, wfs):
    dot_pr_x1x2 = 0.0
    for k in x1.keys():
        dot_pr_x1x2 += np.dot(x1[k].conj(), x2[k]).real
    dot_pr_x1x2 = wfs.kd.comm.sum(dot_pr_x1x2)
    return dot_pr_x1x2


def apply_prec(prec, x, const=1.0):
    y = {}
    for k in x.keys():
        if prec[k].dtype == complex:
            y[k] = const * (prec[k].real * x[k].real +
                            1.0j * prec[k].imag * x[k].imag)
        else:
            y[k] = const * prec[k] * x[k]
    return y



import numpy as np

from gpaw.response import timer
from gpaw.response.kspair import KohnShamKPointPair
from gpaw.response.pair import phase_shifted_fft_indices


class PairDensity:
    """Data class for transition distributed pair density arrays."""

    def __init__(self, tblocks, n_mytG):
        self.tblocks = tblocks
        self.n_mytG = n_mytG

    @classmethod
    def from_qpd(cls, tblocks, qpd):
        n_mytG = qpd.zeros(tblocks.blocksize)
        return cls(tblocks, n_mytG)

    @property
    def local_array_view(self):
        return self.n_mytG[:self.tblocks.nlocal]

    def get_global_array(self):
        """Get the global (all gathered) pair density array n_tG."""
        n_tG = self.tblocks.all_gather(self.n_mytG)

        return n_tG


class NewPairDensityCalculator:
    r"""Class for calculating pair densities

    n_kt(G+q) = n_nks,n'k+qs'(G+q) = <nks| e^-i(G+q)r |n'k+qs'>_V0

                /
              = | dr e^-i(G+q)r ψ_nks^*(r) ψ_n'k+qs'(r)
                /V0

    for a single k-point pair (k, k + q) in the plane-wave mode."""
    def __init__(self, gs, context):
        self.gs = gs
        self.context = context

        # Save PAW correction for all calls with same q_c
        self._pawcorr = None
        self._currentq_c = None

    def initialize_paw_corrections(self, qpd):
        """Initialize the PAW corrections ahead of the actual calculation."""
        self.get_paw_corrections(qpd)

    def get_paw_corrections(self, qpd):
        """Get PAW corrections correcsponding to a specific q-vector."""
        if self._pawcorr is None \
           or not np.allclose(qpd.q_c - self._currentq_c, 0.):
            with self.context.timer('Initialize PAW corrections'):
                self._pawcorr = self.gs.pair_density_paw_corrections(qpd)
                self._currentq_c = qpd.q_c

        return self._pawcorr

    @timer('Calculate pair density')
    def __call__(self, kptpair: KohnShamKPointPair, qpd) -> PairDensity:
        r"""Calculate the pair density for all transitions t.

        In the PAW method, the all-electron pair density is calculated in
        two contributions, the pseudo pair density and a PAW correction,

        n_kt(G+q) = ñ_kt(G+q) + Δn_kt(G+q),

        see [PRB 103, 245110 (2021)] for details.
        """
        # Initialize a blank pair density object
        pair_density = PairDensity.from_qpd(kptpair.tblocks, qpd)
        n_mytG = pair_density.local_array_view

        self.add_pseudo_pair_density(kptpair, qpd, n_mytG)
        self.add_paw_correction(kptpair, qpd, n_mytG)

        return pair_density

    @timer('Calculate the pseudo pair density')
    def add_pseudo_pair_density(self, kptpair, qpd, n_mytG):
        r"""Add the pseudo pair density to an output array.

        The pseudo pair density is first evaluated on the coarse real-space
        grid and then FFT'ed to reciprocal space:

                    /               ˷          ˷
        ñ_kt(G+q) = | dr e^-i(G+q)r ψ_nks^*(r) ψ_n'k+qs'(r)
                    /V0
                                 ˷          ˷
                  = FFT_G[e^-iqr ψ_nks^*(r) ψ_n'k+qs'(r)]
        """
        ikpt1 = kptpair.ikpt1
        ikpt2 = kptpair.ikpt2

        # Map the k-points from the irreducible part of the BZ to the BZ
        # k-point K (up to a reciprocal lattice vector)
        k1_c = self.gs.ibz2bz[kptpair.K1].map_kpoint()
        k2_c = self.gs.ibz2bz[kptpair.K2].map_kpoint()

        # Fourier transform the periodic part of the pseudo waves to the coarse
        # real-space grid and map them to the BZ k-point K (up to the same
        # reciprocal lattice vector as above)
        ut1_hR = self.get_periodic_pseudo_waves(kptpair.K1, ikpt1)
        ut2_hR = self.get_periodic_pseudo_waves(kptpair.K2, ikpt2)

        # Calculate the pseudo pair density in real space, up to a phase of
        # e^(-i[k+q-k']r).
        # This phase does not necessarily vanish, since k2_c only is required
        # to equal k1_c + qpd.q_c modulo a reciprocal lattice vector.
        ut1cc_mytR = ut1_hR[ikpt1.h_myt].conj()
        nt_mytR = ut1cc_mytR * ut2_hR[ikpt2.h_myt]

        # Get the FFT indices corresponding to the Fourier transform
        #                       ˷          ˷
        # FFT_G[e^(-i[k+q-k']r) u_nks^*(r) u_n'k's'(r)]
        Q_G = phase_shifted_fft_indices(k1_c, k2_c, qpd)

        # Add the desired plane-wave components of the FFT'ed pseudo pair
        # density to the output array
        nlocalt = kptpair.tblocks.nlocal
        assert len(n_mytG) == len(nt_mytR) == nlocalt
        for n_G, n_R in zip(n_mytG, nt_mytR):
            n_G[:] += qpd.fft(n_R, 0, Q_G) * qpd.gd.dv

    @timer('Calculate the pair density PAW corrections')
    def add_paw_correction(self, kptpair, qpd, n_mytG):
        r"""Add the pair-density PAW correction to the output array.

        The correction is calculated as a sum over augmentation spheres a
        and projector indices i and j,
                     __  __
                     \   \   ˷     ˷     ˷    ˷
        Δn_kt(G+q) = /   /  <ψ_nks|p_ai><p_aj|ψ_n'k+qs'> Q_aij(G+q)
                     ‾‾  ‾‾
                     a   i,j

        where the pair-density PAW correction tensor is calculated from the
        smooth and all-electron partial waves:

                     /
        Q_aij(G+q) = | dr e^-i(G+q)r [φ_ai^*(r-R_a) φ_aj(r-R_a)
                     /V0                ˷             ˷
                                      - φ_ai^*(r-R_a) φ_aj(r-R_a)]
        """
        ikpt1 = kptpair.ikpt1
        ikpt2 = kptpair.ikpt2

        # Map the projections from the irreducible part of the BZ to the BZ
        # k-point K
        P1h = self.gs.ibz2bz[kptpair.K1].map_projections(ikpt1.Ph)
        P2h = self.gs.ibz2bz[kptpair.K2].map_projections(ikpt2.Ph)

        # Calculate the actual PAW corrections
        Q_aGii = self.get_paw_corrections(qpd).Q_aGii
        P1 = ikpt1.projectors_in_transition_index(P1h)
        P2 = ikpt2.projectors_in_transition_index(P2h)
        for a, Q_Gii in enumerate(Q_aGii):  # Loop over augmentation spheres
            assert P1.atom_partition.comm.size ==\
                P2.atom_partition.comm.size == 1,\
                'We need access to the projections of all atoms'
            P1_myti = P1[a]
            P2_myti = P2[a]
            # Make outer product of the projectors in the projector index i,j
            P1ccP2_mytii = P1_myti.conj()[..., np.newaxis] \
                * P2_myti[:, np.newaxis]
            # Sum over projector indices and add correction to the output
            n_mytG[:] += np.einsum('tij, Gij -> tG', P1ccP2_mytii, Q_Gii)

    def get_periodic_pseudo_waves(self, K, ikpt):
        """FFT the Kohn-Sham orbitals to real space and map them from the
        irreducible k-point to the k-point in question."""
        ut_hR = self.gs.gd.empty(ikpt.nh, self.gs.dtype)
        for h, psit_G in enumerate(ikpt.psit_hG):
            ut_hR[h] = self.gs.ibz2bz[K].map_pseudo_wave(
                self.gs.global_pd.ifft(psit_G, ikpt.ik))

        return ut_hR



import pytest
import numpy as np
from gpaw.xc.libxc import LibXC, short_names
from gpaw.xc.kernel import XCKernel, codes
from gpaw.test import equal


@pytest.mark.ci
@pytest.mark.libxc
def test_xc_xc():
    funcs = []
    modes = []
    for name in short_names:
        try:
            LibXC(name)
        except NameError:
            continue
        if name == 'SCAN':
            continue
        funcs.append(name)
        modes.append(0)
    for name in codes:
        funcs.append(name)
        modes.append(1)

    def create_xc(func, mode):
        if mode == 0:
            xc = LibXC(func)
        else:
            xc = XCKernel(func)
        return xc

    def f1(n_xg, xc):
        e_g = np.empty_like(n_xg[0])
        n_sg = n_xg[:1]
        sigma_xg = n_xg[1:2]
        tau_sg = n_xg[2:]
        dedn_sg = np.zeros_like(n_sg)
        dedsigma_xg = np.zeros_like(sigma_xg)
        dedtau_sg = np.zeros_like(tau_sg)
        xc.calculate(e_g,
                     n_sg, dedn_sg,
                     sigma_xg, dedsigma_xg,
                     tau_sg, dedtau_sg)
        return e_g, np.concatenate((dedn_sg, dedsigma_xg, dedtau_sg))

    def f2(n_xg, xc):
        e_g = np.empty_like(n_xg[0])
        n_sg = n_xg[:2]
        sigma_xg = n_xg[2:5]
        # Triangle inequality:
        assert (sigma_xg[1] <= (sigma_xg[0] + sigma_xg[2]) / 2).all()
        assert (sigma_xg[1] >= -(sigma_xg[0] + sigma_xg[2]) / 2).all()
        tau_sg = n_xg[5:]
        dedn_sg = np.zeros_like(n_sg)
        dedsigma_xg = np.zeros_like(sigma_xg)
        dedtau_sg = np.zeros_like(tau_sg)
        xc.calculate(e_g,
                     n_sg, dedn_sg,
                     sigma_xg, dedsigma_xg,
                     tau_sg, dedtau_sg)
        return e_g, np.concatenate((dedn_sg, dedsigma_xg, dedtau_sg))

    eps = 1.0e-6

    n_xg = np.array(
        [[0.2, 0.01, 0.4],
         [0.2, 0.1, 0.5],
         [0.01, 0.01, 0.2],
         [0.1, 0.3, 0.5]]).T.copy()

    for i, func in enumerate(funcs):
        print(i, func)
        xc = create_xc(funcs[i], modes[i])
        e0_g, d0_xg = f1(n_xg, xc)
        d_xg = np.empty_like(d0_xg)
        for x, n_g in enumerate(n_xg):
            m_xg = n_xg.copy()
            m_xg[x] += eps
            d_xg[x] = 0.5 * f1(m_xg, xc)[0] / eps
            m_xg[x] -= 2 * eps
            d_xg[x] -= 0.5 * f1(m_xg, xc)[0] / eps

        ns_xg = np.empty((7, len(n_g)))
        ns_xg[:2] = n_xg[0] / 2
        ns_xg[2:5] = n_xg[1] / 4
        ns_xg[5:] = n_xg[2] / 2
        es_g, ds_xg = f2(ns_xg, xc)
        error = (abs(d0_xg - d_xg).max() +
                 abs(es_g - e0_g).max() +
                 abs(ds_xg[:2] - d0_xg[0]).max() +
                 abs(ds_xg[2:5].sum(0) / 4 - d0_xg[1]).max() +
                 abs(ds_xg[5:] - d0_xg[2]).max())
        equal(error, 0, 6e-9)
        del xc

    # Numbers from old lxc_xc.py test:
    na = 2.0
    nb = 1.0
    sigma0 = 2.0  # (0.0, 1.0, 1.0)
    sigma1 = 2.0
    sigma2 = 5.0  # (1.0, 2.0, 0.0)
    taua = (3 * np.pi**2)**(2. / 3.) * na**(5. / 3.) / 2 * sigma0
    taub = (3 * np.pi**2)**(2. / 3.) * nb**(5. / 3.) / 2 * sigma2

    n_xg = np.array(
        [[na, nb, sigma0, sigma1, sigma2, taua, taub],
         [0.1, 0.1, 0.025, 0.02, 0.025, 0.25, 0.25],
         [0.1, 0.1, 0.125, 0.12, 0.125, 0.0025, 0.025],
         [0.1, 0.1, 0.01, 0.01, 0.015, 0.2, 0.2],
         [0.1, 0.2, 0.1, -0.08, 0.10, 0.01, 0.05],
         [0.1, 0.1, 0.1, 0.01, 0.01, 0.01, 0.01],
         [0.1, 0.1, 0.1, 0.14, 0.20, 0.01, 0.05]]).T.copy()

    for i, func in enumerate(funcs):
        xc = create_xc(funcs[i], modes[i])
        if xc.type == 'MGGA':
            N_xg = n_xg[:, :1].copy()
        else:
            N_xg = n_xg
        e0_g, d0_xg = f2(N_xg, xc)
        d_xg = np.empty_like(d0_xg)
        for x, n_g in enumerate(N_xg):
            m_xg = N_xg.copy()
            m_xg[x] += eps
            d_xg[x] = 0.5 * f2(m_xg, xc)[0] / eps
            m_xg[x] -= 2 * eps
            d_xg[x] -= 0.5 * f2(m_xg, xc)[0] / eps
        equal(abs(d0_xg - d_xg).max(), 0, 2e-8)
        del xc



import pytest
import numpy as np
from gpaw.mpi import world
from gpaw.blacs import BlacsGrid
from gpaw.blacs import Redistributor


def parallel_eigh(matrixfile, blacsgrid=(4, 2), blocksize=64):
    """Diagonalize matrix in parallel"""
    assert np.prod(blacsgrid) == world.size
    grid = BlacsGrid(world, *blacsgrid)

    if world.rank == 0:
        H_MM = np.load(matrixfile, allow_pickle=True)
        assert H_MM.ndim == 2
        assert H_MM.shape[0] == H_MM.shape[1]
        NM = len(H_MM)
    else:
        NM = 0
    NM = world.sum(NM)  # distribute matrix shape to all nodes

    # descriptor for the individual blocks
    block_desc = grid.new_descriptor(NM, NM, blocksize, blocksize)

    # descriptor for global array on MASTER
    local_desc = grid.new_descriptor(NM, NM, NM, NM)

    # Make some dummy array on all the slaves
    if world.rank != 0:
        H_MM = local_desc.zeros()
    assert local_desc.check(H_MM)

    # The local version of the matrix
    H_mm = block_desc.empty()

    # Distribute global array to smaller blocks
    redistributor = Redistributor(world, local_desc, block_desc)
    redistributor.redistribute(H_MM, H_mm)

    # Allocate arrays for eigenvalues and -vectors
    eps_M = np.empty(NM)
    C_mm = block_desc.empty()
    block_desc.diagonalize_ex(H_mm, C_mm, eps_M)

    # Collect eigenvectors on MASTER
    C_MM = local_desc.empty()
    redistributor2 = Redistributor(world, block_desc, local_desc)
    redistributor2.redistribute(C_mm, C_MM)

    # Return eigenvalues and -vectors on Master
    if world.rank == 0:
        return eps_M, C_MM
    else:
        return None, None


@pytest.mark.intel
def test_parallel_parallel_eigh(in_tmp_dir, scalapack):
    # Test script which should be run on 1, 2, 4, or 8 CPUs

    if world.size == 1:
        blacsgrid = (1, 1)
    elif world.size == 2:
        blacsgrid = (2, 1)
    elif world.size == 4:
        blacsgrid = (2, 2)
    elif world.size == 8:
        blacsgrid = (4, 2)
    else:
        raise RuntimeError('Please use 1, 2, 4, or 8 nodes for this test')

    if world.rank == 0:
        a = np.diag(range(1, 51)).astype(float)
        a.dump('H_50x50.pckl')
        eps, U = np.linalg.eigh(a)

    eps, U = parallel_eigh('H_50x50.pckl', blacsgrid, blocksize=6)
    if world.rank == 0:
        assert abs(eps - range(1, 51)).sum() < 1e-5
        assert abs(U - np.identity(50)).sum() < 1e-5


if __name__ == '__main__':
    test_parallel_parallel_eigh(0)



import pytest
from gpaw.mpi import world
from gpaw import GPAW, ConvergenceError, restart
from gpaw.mixer import MixerSum
from ase.build import bulk

# bulk Fe with k-point, band, and domain parallelization
pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


def test_fileio_parallel(in_tmp_dir):
    a = 2.87
    atoms = bulk('Fe', 'bcc', a=a)
    atoms.set_initial_magnetic_moments([2.2])
    calc = GPAW(h=0.20,
                experimental={'niter_fixdensity': 2},
                eigensolver='rmm-diis',
                mixer=MixerSum(0.1, 3),
                nbands=6,
                kpts=(4, 4, 4),
                parallel={'band': 2, 'domain': (2, 1, 1)},
                maxiter=4)
    atoms.calc = calc
    try:
        atoms.get_potential_energy()
    except ConvergenceError:
        pass
    calc.write('tmp.gpw', mode='all')

    # Continue calculation for few iterations
    atoms, calc = restart('tmp.gpw',
                          eigensolver='rmm-diis',
                          mixer=MixerSum(0.1, 3),
                          parallel={'band': 2, 'domain': (1, 1, 2)},
                          maxiter=4)
    try:
        atoms.get_potential_energy()
    except ConvergenceError:
        pass



"""Helper classes for doing jellium calculations."""
from math import pi

import numpy as np
from ase.units import Bohr


def create_background_charge(**kwargs):
    if 'z1' in kwargs:
        return JelliumSlab(**kwargs)
    return Jellium(**kwargs)


class Jellium():
    """ The Jellium object """
    def __init__(self, charge):
        """ Initialize the Jellium object

        Input: charge, the total Jellium background charge.
        """
        self.charge = charge
        self.rs = None  # the Wigner-Seitz radius
        self.volume = None
        self.mask_g = None
        self.gd = None

    def todict(self):
        return {'charge': self.charge}

    def set_grid_descriptor(self, gd):
        """ Set the grid descriptor for the Jellium background charge"""
        self.gd = gd
        self.mask_g = self.get_mask().astype(float)
        self.volume = self.gd.comm.sum(self.mask_g.sum()) * self.gd.dv
        if self.charge != 0.:
            self.rs = (3 / pi / 4 * self.volume / abs(self.charge))**(1 / 3)
        else:
            self.rs = np.inf  # Avoid having to see numpy's warning.

    def get_mask(self):
        """Choose which grid points are inside the jellium.

        gd: grid descriptor

        Return ndarray of ones and zeros indicating where the jellium
        is.  This implementation will put the positive background in the
        whole cell.  Overwrite this method in subclasses."""

        return self.gd.zeros() + 1.0

    def add_charge_to(self, rhot_g):
        """ Add Jellium background charge to pseudo charge density rhot_g"""
        rhot_g -= self.mask_g * (self.charge / self.volume)

    def add_fourier_space_charge_to(self, pd, rhot_q):
        rhot_g = pd.gd.zeros()
        self.add_charge_to(rhot_g)
        rhot_q += pd.fft(rhot_g)


class JelliumSlab(Jellium):
    """ The Jellium slab object """
    def __init__(self, charge, z1, z2):
        """Put the positive background charge where z1 < z < z2.

        z1: float
            Position of lower surface in Angstrom units.
        z2: float
            Position of upper surface in Angstrom units."""
        Jellium.__init__(self, charge)
        self.z1 = (z1 - 0.0001) / Bohr
        self.z2 = (z2 - 0.0001) / Bohr

    def todict(self):
        dct = Jellium.todict(self)
        dct.update(z1=self.z1 * Bohr + 0.0001,
                   z2=self.z2 * Bohr + 0.0001)
        return dct

    def get_mask(self):
        r_gv = self.gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        # r_gv: 4-dimensional ndarray
        # positions of the grid points in Bohr units.
        return np.logical_and(r_gv[:, :, :, 2] > self.z1,
                              r_gv[:, :, :, 2] < self.z2)



from ase import Atoms

from gpaw import GPAW, FermiDirac
from gpaw.test import equal


def test_spin_spinFe3plus():
    h = 0.4
    q = 3

    s = Atoms('Fe')
    s.center(vacuum=2.5)
    convergence = {'eigenstates': 0.01, 'density': 0.1, 'energy': 0.1}

    # use Hunds rules

    c = GPAW(charge=q, h=h, nbands=5,
             hund=True,
             eigensolver='rmm-diis',
             occupations=FermiDirac(width=0.1),
             convergence=convergence)
    s.calc = c
    equal(s.get_magnetic_moment(), 5, 0.1)

    # set magnetic moment

    mm = [5]
    s.set_initial_magnetic_moments(mm)
    c = GPAW(charge=q, h=h, nbands=5,
             occupations=FermiDirac(width=0.1, fixmagmom=True),
             convergence=convergence)
    s.calc = c
    equal(s.get_magnetic_moment(), 5, 0.1)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import numpy as np

from gpaw import debug
from gpaw.utilities import is_contiguous
import _gpaw


class Spline:
    def __init__(self, l, rmax, f_g):
        """Spline object

        The integer l gives the angular momentum quantum number and
        the list contains the spline values from r=0 to r=rcut.

        The array f_g gives the radial part of the function on the grid.
        The radial function is multiplied by a real solid spherical harmonics
        (r^l * Y_lm).
        """
        assert rmax > 0.0
        f_g = np.array(f_g, float)
        # Copy so we don't change the values of the input array
        f_g[-1] = 0.0
        self.spline = _gpaw.Spline(l, rmax, f_g)
        self.l = l

    def get_cutoff(self):
        """Return the radial cutoff."""
        return self.spline.get_cutoff()

    def get_angular_momentum_number(self):
        """Return the angular momentum quantum number."""
        return self.spline.get_angular_momentum_number()

    def __repr__(self):
        return ('Spline(l={}, rmax={:.2f}, ...)'
                .format(self.get_angular_momentum_number(),
                        self.get_cutoff()))

    def get_value_and_derivative(self, r):
        """Return the value and derivative."""
        return self.spline.get_value_and_derivative(r)

    def __call__(self, r):
        assert r >= 0.0
        return self.spline(r)

    def map(self, r_x):
        return np.vectorize(self, [float])(r_x)

    def get_functions(self, gd, start_c, end_c, spos_c):
        h_cv = gd.h_cv
        # start_c is the new origin so we translate gd.beg_c to start_c
        origin_c = np.array([0, 0, 0])
        pos_v = np.dot(spos_c, gd.cell_cv) - np.dot(start_c, h_cv)
        A_gm, G_b = _gpaw.spline_to_grid(self.spline,
                                         origin_c,
                                         end_c - start_c,
                                         pos_v,
                                         h_cv,
                                         end_c - start_c,
                                         origin_c)

        if debug:
            assert G_b.ndim == 1 and G_b.shape[0] % 2 == 0
            assert is_contiguous(G_b, np.intc)
            assert A_gm.shape[:-1] == np.sum(G_b[1::2] - G_b[::2])

        indices_gm, ng, nm = self.spline.get_indices_from_zranges(start_c,
                                                                  end_c, G_b)
        shape = (nm,) + tuple(end_c - start_c)
        work_mB = np.zeros(shape, dtype=A_gm.dtype)
        np.put(work_mB, indices_gm, A_gm)
        return work_mB



"""Symmetry checking code."""
import sys
from typing import Any, Dict, List, Sequence, Union

import numpy as np
from ase import Atoms
from numpy.linalg import det, inv, solve
from scipy.ndimage import map_coordinates

from gpaw.typing import Array1D, Array2D, Array3D, ArrayLike

from . import PointGroup

Axis = Union[str, Sequence[float], Array1D, None]


class SymmetryChecker:
    def __init__(self,
                 group: Union[str, PointGroup],
                 center: ArrayLike,
                 radius: float = 2.0,
                 x: Axis = None,
                 y: Axis = None,
                 z: Axis = None,
                 grid_spacing: float = 0.2):
        """Check point-group symmetries.

        If a non-standard orientation is desired then two of
        *x*, *y*, *z* can be specified.
        """
        if isinstance(group, str):
            group = PointGroup(group)
        self.group = group
        self.normalized_table = group.get_normalized_table()
        self.points = sphere(radius, grid_spacing)
        self.center = center
        self.grid_spacing = grid_spacing
        self.rotation = rotation_matrix([x, y, z])

    def check_atoms(self, atoms: Atoms, tol: float = 1e-5) -> bool:
        """Check if atoms have all the symmetries.

        Unit of *tol* is Angstrom.
        """
        numbers = atoms.numbers
        positions = (atoms.positions - self.center).dot(self.rotation.T)
        icell = np.linalg.inv(atoms.cell.dot(self.rotation.T))
        for opname, op in self.group.operations.items():
            P = positions.dot(op.T)
            for i, pos in enumerate(P):
                sdiff = (pos - positions).dot(icell)
                sdiff -= sdiff.round() * atoms.pbc
                dist2 = (sdiff.dot(atoms.cell)**2).sum(1)
                j = dist2.argmin()
                if dist2[j] > tol**2 or numbers[j] != numbers[i]:
                    return False
        return True

    def check_function(self,
                       function: Array3D,
                       grid_vectors: Array2D = None) -> Dict[str, Any]:
        """Check function on uniform grid."""
        if grid_vectors is None:
            grid_vectors = np.eye(3)
        dv = abs(det(grid_vectors))
        norm1 = (function**2).sum() * dv
        M = inv(grid_vectors).T
        overlaps: List[float] = []
        for op in self.group.operations.values():
            op = self.rotation.T @ op @ self.rotation
            pts = (self.points @ op.T + self.center) @ M.T
            pts %= function.shape
            values = map_coordinates(function, pts.T, mode='wrap')
            if not overlaps:
                values1 = values
            overlaps.append(values.dot(values1) * self.grid_spacing**3)

        reduced_overlaps = []
        i1 = 0
        for n in self.group.nops:
            i2 = i1 + n
            reduced_overlaps.append(sum(overlaps[i1:i2]) / n / overlaps[0])
            i1 = i2

        characters = solve(self.normalized_table.T, reduced_overlaps)
        best = self.group.symmetries[characters.argmax()]

        return {'symmetry': best,
                'norm': norm1,
                'overlaps': overlaps,
                'characters': {symmetry: value
                               for symmetry, value
                               in zip(self.group.symmetries, characters)}}

    def check_band(self,
                   calc,
                   band: int,
                   spin: int = 0) -> Dict[str, Any]:
        """Check wave function from GPAW calculation."""
        wfs = calc.get_pseudo_wave_function(band, spin=spin)
        grid_vectors = (calc.atoms.cell.T / wfs.shape).T
        return self.check_function(wfs, grid_vectors)

    def check_calculation(self,
                          calc,
                          n1: int,
                          n2: int,
                          spin: int = 0,
                          output: str = '-') -> None:
        """Check several wave functions from GPAW calculation."""
        lines = ['band    energy     norm  normcut     best     ' +
                 ''.join(f'{sym:8}' for sym in self.group.symmetries)]
        n2 = n2 or calc.get_number_of_bands()
        for n in range(n1, n2):
            dct = self.check_band(calc, n, spin)
            best = dct['symmetry']
            norm = dct['norm']
            normcut = dct['overlaps'][0]
            eig = calc.get_eigenvalues(spin=spin)[n]
            lines.append(
                f'{n:4} {eig:9.3f} {norm:8.3f} {normcut:8.3f} {best:>8}' +
                ''.join(f'{x:8.3f}'
                        for x in dct['characters'].values()))

        fd = sys.stdout if output == '-' else open(output, 'w')
        fd.write('\n'.join(lines) + '\n')
        if output != '-':
            fd.close()


def sphere(radius: float, grid_spacing: float) -> Array2D:
    """Return sphere of grid-points.

    >>> points = sphere(1.1, 1.0)
    >>> points.shape
    (7, 3)
    """
    npts = int(radius / grid_spacing) + 1
    x = np.linspace(-npts, npts, 2 * npts + 1) * grid_spacing
    points = np.array(np.meshgrid(x, x, x, indexing='ij')).reshape((3, -1)).T
    points = points[(points**2).sum(1) <= radius**2]
    return points


def rotation_matrix(axes: Sequence[Axis]) -> Array3D:
    """Calculate rotation matrix.

    >>> rotation_matrix(['-y', 'x', None])
    array([[ 0, -1,  0],
           [ 1,  0,  0],
           [ 0,  0,  1]])
    """
    if all(axis is None for axis in axes):
        return np.eye(3)

    j = -1
    for i, axis in enumerate(axes):
        if axis is None:
            assert j == -1
            j = i
    assert j != -1

    axes = [normalize(axis) if axis is not None else None
            for axis in axes]
    axes[j] = np.cross(axes[j - 2], axes[j - 1])  # type: ignore

    return np.array(axes)


def normalize(vector: Union[str, Sequence[float], Array1D]) -> Array1D:
    """Normalize a vector.

    The *vector* must be a sequence of three numbers or one of the following
    strings: x, y, z, -x, -y, -z.
    """
    if isinstance(vector, str):
        if vector[0] == '-':
            return -np.array(normalize(vector[1:]))
        return {'x': np.array([1, 0, 0]),
                'y': np.array([0, 1, 0]),
                'z': np.array([0, 0, 1])}[vector]
    return np.array(vector) / np.linalg.norm(vector)



import pickle
import numpy as np
from os.path import isfile

from ase.parallel import parprint
from ase.units import Bohr, Ha

from gpaw.mpi import world
from gpaw.utilities import unpack2
from gpaw.lcao.projected_wannier import dots
from gpaw.utilities.tools import tri2full
from gpaw.lfc import LocalizedFunctionsCollection as LFC


"""This module is used to calculate the electron-phonon coupling matrix,
    expressed in terms of GPAW LCAO orbitals.

    This module is not maintained and possibly broken.
    Use gpaw/elph/electronphonon instead.
    """


class ElectronPhononCouplingMatrix:
    r"""Class for calculating the electron-phonon coupling matrix, defined
    by the electron phonon interaction.

    ::

                  __                   _____
                  \     l   cc        /  h             cc
        H      =   )   M   c   c     /------   ( b  + b   ),
         el-ph    /_    ij  i   j  \/   2 W       l    l
                    l,ij                     l

    where the electron phonon coupling matrix is given by::

            l           ___
            M   = < i | \ /  V   * v  |j>
             ij          'u   eff   l

    """

    def __init__(self, atoms, indices=None, name='v', delta=0.005, nfree=2,
                 derivativemethod='tci'):
        assert nfree in [2, 4]
        self.nfree = nfree
        self.delta = delta

        if indices is None:
            indices = range(len(self.atoms))
        self.calc = atoms.calc
        self.atoms = atoms
        self.indices = np.asarray(indices)
        self.name = name
        self.p0 = self.atoms.positions.copy()

        self.derivativemethod = derivativemethod
        if derivativemethod == 'grid':
            self.get_dP_aMix = get_grid_dP_aMix
        elif derivativemethod == 'grid2':
            self.get_dP_aMix = get_grid2_dP_aMix
        elif derivativemethod == 'tci':
            self.get_dP_aMix = get_tci_dP_aMix
        else:
            raise ValueError('derivativemethod must be grid, grid2, or tci')

    def run(self):
        if not isfile(self.name + '.eq.pckl'):

            self.calc.calculate(self.atoms)
            Vt_G = self.calc.get_effective_potential(pad=False)
            Vt_G = self.calc.wfs.gd.collect(Vt_G, broadcast=True) / Ha
            dH_asp = self.calc.hamiltonian.dH_asp
            setups = self.calc.wfs.setups
            nspins = self.calc.wfs.nspins
            gd_comm = self.calc.wfs.gd.comm

            alldH_asp = {}
            for a, setup in enumerate(setups):
                ni = setup.ni
                nii = ni * (ni + 1) // 2
                tmpdH_sp = np.zeros((nspins, nii))
                if a in dH_asp:
                    tmpdH_sp[:] = dH_asp[a]
                gd_comm.sum(tmpdH_sp)
                alldH_asp[a] = tmpdH_sp

            forces = self.atoms.get_forces()
            self.calc.write('eq.gpw')

            world.barrier()
            if world.rank == 0:
                vd = open(self.name + '.eq.pckl', 'wb')
                fd = open('vib.eq.pckl', 'wb')
                pickle.dump((Vt_G, alldH_asp), vd, 2)
                pickle.dump(forces, fd)
                vd.close()
                fd.close()
            world.barrier()

        p = self.atoms.positions.copy()
        for a in self.indices:
            for j in range(3):
                for sign in [-1, 1]:
                    for ndis in range(1, self.nfree / 2 + 1):
                        name = '.%d%s%s.pckl' % (a, 'xyz'[j],
                                                 ndis * ' +-'[sign])
                        if isfile(self.name + name):
                            continue
                        self.atoms.positions[a, j] = (p[a, j] +
                                                      sign * ndis * self.delta)
                        self.calc.calculate(self.atoms)
                        Vt_G = self.calc.get_effective_potential(pad=False)
                        Vt_G = self.calc.wfs.gd.collect(Vt_G,
                                                        broadcast=True) / Ha
                        dH_asp = self.calc.hamiltonian.dH_asp

                        alldH_asp = {}
                        for a2, setup in enumerate(setups):
                            ni = setup.ni
                            nii = ni * (ni + 1) // 2
                            tmpdH_sp = np.zeros((nspins, nii))
                            if a2 in dH_asp:
                                tmpdH_sp[:] = dH_asp[a2]
                            gd_comm.sum(tmpdH_sp)
                            alldH_asp[a2] = tmpdH_sp

                        forces = self.atoms.get_forces()
                        world.barrier()
                        if world.rank == 0:
                            vd = open(self.name + name, 'wb')
                            fd = open('vib' + name, 'wb')
                            pickle.dump((Vt_G, alldH_asp), vd)
                            pickle.dump(forces, fd)
                            vd.close()
                            fd.close()
                        world.barrier()
                        self.atoms.positions[a, j] = p[a, j]
        self.atoms.set_positions(p)

    def get_gradient(self):
        """Calculates gradient"""
        nx = len(self.indices) * 3
        veqt_G, dHeq_asp = pickle.load(open(self.name + '.eq.pckl', 'rb'))
        gpts = veqt_G.shape
        dvt_Gx = np.zeros(gpts + (nx, ))
        ddH_aspx = {}
        for a, dH_sp in dHeq_asp.items():
            ddH_aspx[a] = np.empty(dH_sp.shape + (nx,))

        x = 0
        for a in self.indices:
            for i in range(3):
                name = '%s.%d%s' % (self.name, a, 'xyz'[i])
                vtm_G, dHm_asp = pickle.load(open(name + '-.pckl', 'rb'))
                vtp_G, dHp_asp = pickle.load(open(name + '+.pckl', 'rb'))

                if self.nfree == 4:
                    vtmm_G, dHmm_asp = pickle.load(
                        open(name + '--.pckl', 'rb'))
                    vtpp_G, dHpp_asp = pickle.load(
                        open(name + '++.pckl', 'rb'))
                    dvtdx_G = (-vtpp_G + 8.0 * vtp_G -
                               8 * vtm_G + vtmm_G) / (12 * self.delta / Bohr)
                    dvt_Gx[:, :, :, x] = dvtdx_G
                    for atom, ddH_spx in ddH_aspx.items():
                        ddH_aspx[atom][:, :, x] = (
                            -dHpp_asp[atom]
                            + 8.0 * dHp_asp[atom]
                            - 8.0 * dHm_asp[atom]
                            + dHmm_asp[atom]) / (12 * self.delta / Bohr)
                else:  # nfree = 2
                    dvtdx_G = (vtp_G - vtm_G) / (2 * self.delta / Bohr)
                    dvt_Gx[..., x] = dvtdx_G
                    for atom, ddH_spx in ddH_aspx.items():
                        ddH_aspx[atom][:, :, x] = (
                            dHp_asp[atom]
                            - dHm_asp[atom]) / (2 * self.delta / Bohr)
                x += 1
        return dvt_Gx, ddH_aspx

    def get_M(self, modes, q=0):
        r"""Calculate el-ph coupling matrix for given modes(s).

        XXX:
        kwarg "q=0" is an ugly hack for k-points.
        There shuold be loops over q!

        Note that modes must be given as a dictionary with mode
        frequencies in eV and corresponding mode vectors in units
        of 1/sqrt(amu), where amu = 1.6605402e-27 Kg is an atomic mass unit.
        In short frequencies and mode vectors must be given in ase units.

        ::

                  d                   d  ~
            < w | -- v | w' > = < w | -- v | w'>
                  dP                  dP

                               _
                              \        ~a     d   .       ~a
                            +  ) < w | p  >   -- /_\H   < p | w' >
                              /_        i     dP     ij    j
                              a,ij

                               _
                              \        d  ~a     .        ~a
                            +  ) < w | -- p  >  /_\H    < p | w' >
                              /_       dP  i        ij     j
                              a,ij

                               _
                              \        ~a     .        d  ~a
                            +  ) < w | p  >  /_\H    < -- p  | w' >
                              /_        i        ij    dP  j
                              a,ij

        """

        modes1 = modes.copy()
        # convert to atomic units
        amu = 1.6605402e-27  # atomic unit mass [Kg]
        me = 9.1093897e-31  # electron mass    [Kg]
        modes = {}
        for k in modes1.keys():
            modes[k / Ha] = modes1[k] / np.sqrt(amu / me)

        dvt_Gx, ddH_aspx = self.get_gradient()

        from gpaw import restart
        atoms, calc = restart('eq.gpw', txt=None)
        if calc.wfs.S_qMM is None:
            calc.initialize(atoms)
            calc.initialize_positions(atoms)

        wfs = calc.wfs
        nao = wfs.setups.nao
        bfs = wfs.basis_functions
        dtype = wfs.dtype
        spin = 0  # XXX

        M_lii = {}
        parprint('Starting gradient of pseudo part')
        for f, mode in modes.items():
            mo = []
            M_ii = np.zeros((nao, nao), dtype)
            for a in self.indices:
                mo.append(mode[a])
            mode = np.asarray(mo).flatten()
            dvtdP_G = np.dot(dvt_Gx, mode)
            bfs.calculate_potential_matrix(dvtdP_G, M_ii, q=q)
            tri2full(M_ii, 'L')
            M_lii[f] = M_ii
        parprint('Finished gradient of pseudo part')

        P_aqMi = calc.wfs.P_aqMi
        # Add the term
        #  _
        # \        ~a     d   .       ~a
        #  ) < w | p  >   -- /_\H   < p | w' >
        # /_        i     dP     ij    j
        # a,ij

        Ma_lii = {}
        for f, mode in modes.items():
            Ma_lii[f] = np.zeros_like(M_lii.values()[0])

        parprint('Starting gradient of dH^a part')
        for f, mode in modes.items():
            mo = []
            for a in self.indices:
                mo.append(mode[a])
            mode = np.asarray(mo).flatten()

            for a, ddH_spx in ddH_aspx.items():
                ddHdP_sp = np.dot(ddH_spx, mode)
                ddHdP_ii = unpack2(ddHdP_sp[spin])
                Ma_lii[f] += dots(P_aqMi[a][q], ddHdP_ii, P_aqMi[a][q].T)
        parprint('Finished gradient of dH^a part')

        parprint('Starting gradient of projectors part')
        dP_aMix = self.get_dP_aMix(calc.spos_ac, wfs, q)
        parprint('Finished gradient of projectors part')

        dH_asp = pickle.load(open('v.eq.pckl', 'rb'))[1]

        Mb_lii = {}
        for f, mode in modes.items():
            Mb_lii[f] = np.zeros_like(M_lii.values()[0])

        for f, mode in modes.items():
            for a, dP_Mix in dP_aMix.items():
                dPdP_Mi = np.dot(dP_Mix, mode[a])
                dH_ii = unpack2(dH_asp[a][spin])
                dPdP_MM = dots(dPdP_Mi, dH_ii, P_aqMi[a][q].T)
                Mb_lii[f] -= dPdP_MM + dPdP_MM.T
                # XXX The minus sign here is quite subtle.
                # It is related to how the derivative of projector
                # functions in GPAW is calculated.
                # More thorough explanations, anyone...?

        # Units of M_lii are Hartree/(Bohr * sqrt(m_e))
        for mode in M_lii.keys():
            M_lii[mode] += Ma_lii[mode] + Mb_lii[mode]

        # conversion to eV. The prefactor 1 / sqrt(hb^2 / 2 * hb * f)
        # has units Bohr * sqrt(me)
        M_lii_1 = M_lii.copy()
        M_lii = {}

        for f in M_lii_1.keys():
            M_lii[f * Ha] = M_lii_1[f] * Ha / np.sqrt(2 * f)

        return M_lii


#####################################################
# XXX grid and grid 2 sometimes gives random numbers,
# XXX sometimes even nan!
#####################################################

def get_grid_dP_aMix(spos_ac, wfs, q):  # XXXXXX q
    nao = wfs.setups.nao
    C_MM = np.identity(nao, dtype=wfs.dtype)
    # XXX In the future use the New Two-Center integrals
    # to evaluate this
    dP_aMix = {}
    for a, setup in enumerate(wfs.setups):
        ni = 0
        dP_Mix = np.zeros((nao, setup.ni, 3))
        pt = LFC(wfs.gd, [setup.pt_j],
                 wfs.kd.comm, dtype=wfs.dtype, forces=True)
        spos1_ac = [spos_ac[a]]
        pt.set_k_points(wfs.ibzk_qc)
        pt.set_positions(spos1_ac)
        for b, setup_b in enumerate(wfs.setups):
            nao = setup_b.nao
            phi_MG = wfs.gd.zeros(nao, wfs.dtype)
            phi_MG = wfs.gd.collect(phi_MG, broadcast=False)
            wfs.basis_functions.lcao_to_grid(C_MM[ni:ni + nao], phi_MG, q)
            dP_bMix = pt.dict(len(phi_MG), derivative=True)
            pt.derivative(phi_MG, dP_bMix, q=q)
            dP_Mix[ni:ni + nao] = dP_bMix[0]
            ni += nao
            parprint('projector grad. doing atoms (%s, %s) ' % (a, b))

        dP_aMix[a] = dP_Mix
    return dP_aMix


def get_grid2_dP_aMix(spos_ac, wfs, q, *args, **kwargs):  # XXXXXX q
    nao = wfs.setups.nao
    C_MM = np.identity(nao, dtype=wfs.dtype)
    bfs = wfs.basis_functions
    phi_MG = wfs.gd.zeros(nao, wfs.dtype)
    bfs.lcao_to_grid(C_MM, phi_MG, q)
    setups = wfs.setups
    pt = LFC(wfs.gd, [setup.pt_j for setup in setups],
             wfs.kd.comm, dtype=wfs.dtype, forces=True)
    pt.set_k_points(wfs.ibzk_qc)
    pt.set_positions(spos_ac)
    dP_aMix = pt.dict(len(phi_MG), derivative=True)
    pt.derivative(phi_MG, dP_aMix, q=q)
    return dP_aMix


def get_tci_dP_aMix(spos_ac, wfs, q, *args, **kwargs):
    # container for spline expansions of basis function-projector pairs
    # (note to self: remember to conjugate/negate because of that)
    from gpaw.lcao.overlap import ManySiteDictionaryWrapper,\
        TwoCenterIntegralCalculator, NewTwoCenterIntegrals

    if not isinstance(wfs.tci, NewTwoCenterIntegrals):
        raise RuntimeError('Please remember --gpaw=usenewtci=True')

    dP_aqxMi = {}
    nao = wfs.setups.nao
    nq = len(wfs.ibzk_qc)
    for a, setup in enumerate(wfs.setups):
        dP_aqxMi[a] = np.zeros((nq, 3, nao, setup.ni), wfs.dtype)

    calc = TwoCenterIntegralCalculator(wfs.ibzk_qc, derivative=True)
    expansions = ManySiteDictionaryWrapper(wfs.tci.P_expansions, dP_aqxMi)
    calc.calculate(wfs.tci.atompairs, [expansions], [dP_aqxMi])

    dP_aMix = {}
    for a in dP_aqxMi:
        dP_aMix[a] = dP_aqxMi[a].transpose(0, 2, 3, 1).copy()[q]  # XXX q
    return dP_aMix



from ase import Atom, Atoms

from gpaw import GPAW, Mixer, RMMDIIS
from gpaw.test import equal


def test_eigen_blocked_rmm_diis(in_tmp_dir):
    a = 4.0
    n = 20
    d = 1.0
    x = d / 3**0.5
    atoms = Atoms([Atom('C', (0.0, 0.0, 0.0)),
                   Atom('H', (x, x, x)),
                   Atom('H', (-x, -x, x)),
                   Atom('H', (x, -x, -x)),
                   Atom('H', (-x, x, -x))],
                  cell=(a, a, a), pbc=True)
    calc = GPAW(gpts=(n, n, n), nbands=4, txt='a.txt',
                mixer=Mixer(0.25, 3, 1), eigensolver='rmm-diis')
    atoms.calc = calc
    e0 = atoms.get_potential_energy()
    niter0 = calc.get_number_of_iterations()

    es = RMMDIIS(blocksize=3)
    calc = GPAW(gpts=(n, n, n), nbands=4, txt='b.txt',
                mixer=Mixer(0.25, 3, 1), eigensolver=es)
    atoms.calc = calc
    e1 = atoms.get_potential_energy()
    niter1 = calc.get_number_of_iterations()
    equal(e0, e1, 0.000001)
    equal(niter0, niter1, 0)



from __future__ import annotations

from gpaw.core import UniformGrid
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.fd_operators import Laplace
from gpaw.new import zip
from gpaw.new.builder import create_uniform_grid
from gpaw.new.fd.pot_calc import UniformGridPotentialCalculator
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.poisson import PoissonSolver, PoissonSolverWrapper
from gpaw.new.pwfd.builder import PWFDDFTComponentsBuilder
from gpaw.poisson import PoissonSolver as make_poisson_solver


class FDDFTComponentsBuilder(PWFDDFTComponentsBuilder):
    def __init__(self, atoms, params, nn=3, interpolation=3):
        super().__init__(atoms, params)
        assert not self.soc
        self.kin_stencil_range = nn
        self.interpolation_stencil_range = interpolation

        self._nct_aR = None

    def create_uniform_grids(self):
        grid = create_uniform_grid(
            self.mode,
            self.params.gpts,
            self.atoms.cell,
            self.atoms.pbc,
            self.ibz.symmetries,
            h=self.params.h,
            interpolation='not fft',
            comm=self.communicators['d'])
        fine_grid = grid.new(size=grid.size_c * 2)
        # decomposition=[2 * d for d in grid.decomposition]
        return grid, fine_grid

    def create_wf_description(self) -> UniformGrid:
        return self.grid.new(dtype=self.dtype)

    def get_pseudo_core_densities(self):
        if self._nct_aR is None:
            self._nct_aR = self.setups.create_pseudo_core_densities(
                self.grid, self.fracpos_ac, atomdist=self.atomdist)
        return self._nct_aR

    def create_poisson_solver(self) -> PoissonSolver:
        solver = make_poisson_solver(**self.params.poissonsolver)
        solver.set_grid_descriptor(self.fine_grid._gd)
        return PoissonSolverWrapper(solver)

    def create_potential_calculator(self):
        poisson_solver = self.create_poisson_solver()
        nct_aR = self.get_pseudo_core_densities()
        return UniformGridPotentialCalculator(self.grid,
                                              self.fine_grid,
                                              self.setups,
                                              self.xc, poisson_solver,
                                              nct_aR, self.nct_R,
                                              self.interpolation_stencil_range)

    def create_hamiltonian_operator(self, blocksize=10):
        return FDHamiltonian(self.wf_desc, self.kin_stencil_range, blocksize)

    def convert_wave_functions_from_uniform_grid(self,
                                                 C_nM,
                                                 basis_set,
                                                 kpt_c,
                                                 q):
        grid = self.grid.new(kpt=kpt_c, dtype=self.dtype)
        psit_nR = grid.zeros(self.nbands, self.communicators['b'])
        mynbands = len(C_nM.data)
        basis_set.lcao_to_grid(C_nM.data, psit_nR.data[:mynbands], q)
        return psit_nR

    def read_ibz_wave_functions(self, reader):
        ibzwfs = super().read_ibz_wave_functions(reader)

        if 'coefficients' in reader.wave_functions:
            name = 'coefficients'
        elif 'values' in reader.wave_functions:
            name = 'values'
        else:
            return ibzwfs

        c = reader.bohr**1.5
        if reader.version < 0:
            c = 1  # old gpw file

        for wfs in ibzwfs:
            grid = self.wf_desc.new(kpt=wfs.kpt_c)
            index = (wfs.spin, wfs.k)
            data = reader.wave_functions.proxy(name, *index)
            data.scale = c
            if self.communicators['w'].size == 1:
                wfs.psit_nX = UniformGridFunctions(grid, self.nbands,
                                                   data=data)
            else:
                band_comm = self.communicators['b']
                wfs.psit_nX = UniformGridFunctions(
                    grid, self.nbands,
                    comm=band_comm)
                if grid.comm.rank == 0:
                    mynbands = (self.nbands +
                                band_comm.size - 1) // band_comm.size
                    n1 = min(band_comm.rank * mynbands, self.nbands)
                    n2 = min((band_comm.rank + 1) * mynbands, self.nbands)
                    assert wfs.psit_nX.mydims[0] == n2 - n1
                    data = data[n1:n2]  # read from file
                wfs.psit_nX.scatter_from(data)

        return ibzwfs


class FDHamiltonian(Hamiltonian):
    def __init__(self, grid, kin_stencil=3, blocksize=10):
        self.grid = grid
        self.blocksize = blocksize
        self._gd = grid._gd
        self.kin = Laplace(self._gd, -0.5, kin_stencil, grid.dtype)

    def apply(self, vt_sR, psit_nR, out, spin):
        self.kin.apply(psit_nR.data, out.data, psit_nR.desc.phase_factors_cd)
        for p, o in zip(psit_nR.data, out.data):
            o += p * vt_sR.data[spin]
        return out

    def create_preconditioner(self, blocksize):
        from types import SimpleNamespace

        from gpaw.preconditioner import Preconditioner as PC
        pc = PC(self._gd, self.kin, self.grid.dtype, self.blocksize)

        def apply(psit, residuals, out):
            kpt = SimpleNamespace(phase_cd=psit.desc.phase_factors_cd)
            pc(residuals.data, kpt, out=out.data)

        return apply



from __future__ import annotations

import numpy as np
from gpaw.core.domain import Domain
from gpaw.typing import Vector


class SpinorWaveFunctionDescriptor(Domain):
    def __init__(self,
                 desc: Domain,
                 qspiral_v: Vector = None):
        self.desc = desc
        self.qspiral_v = (np.asarray(qspiral_v) if qspiral_v is not None else
                          None)
        Domain.__init__(self, desc.cell_cv, desc.pbc_c, desc.kpt_c, desc.comm,
                        complex)
        self.myshape = (2,) + desc.myshape
        self.itemsize = desc.itemsize

    def __repr__(self):
        return f'{self.__class__.__name__}({self.desc}, {self.qspiral_v})'

    def new(self, *, kpt):
        desc = self.desc.new(kpt=kpt)
        desc.qspiral_v = self.qspiral_v
        return SpinorWaveFunctionDescriptor(desc, self.qspiral_v)

    def empty(self, nbands, band_comm, xp=None):
        return self.desc.empty((nbands, 2), band_comm)

    def global_shape(self) -> tuple[int, ...]:
        return (2,) + self.desc.global_shape()
